from pygbx.headers import CGameHeader, CGameCtnCollectorList, CollectorStock, MapBlock, Vector3, CGameChallenge, CGameBlockItem, CGameWaypointSpecialProperty, CGameCommon, CGameReplayRecord, CGameGhost, CGameCtnGhost, ControlEntry, GhostSampleRecord
from pygbx.bytereader import ByteReader
from pygbx.stadium_blocks import STADIUM_BLOCKS
from pygbx.canyon_blocks import CANYON_BLOCKS
from pygbx.valley_blocks import VALLEY_BLOCKS
from pygbx.lagoon_blocks import LAGOON_BLOCKS
from pygbx.stadium_block_offsets import STADIUM_BLOCK_OFFSETS
from pygbx.canyon_block_offsets import CANYON_BLOCK_OFFSETS
from pygbx.gbx import Gbx, GbxType, GbxLoadError


CANYON_BLOCKS = {
    'RoadRaceStraight': 1,
    'RoadRaceStraightScreen': 2,
    'RoadRaceStraightOffice': 3,
    'RoadRaceStraightWindows': 4,
    'RoadRaceStraightJump': 5,
    'RoadRaceSwitchStraight1x1Mirror': 6,
    'RoadRaceSwitchStraight1x1': 7,
    'RoadRaceCornerIn': 8,
    'RoadRaceCornerInScreen': 9,
    'RoadRaceCornerOut': 10,
    'RoadRaceGTCurve2In': 11,
    'RoadRaceGTCurve2Out': 12,
    'RoadRaceSlopeBaseGTCurve2In': 13,
    'RoadRaceSlopeBaseGTCurve2InMirror': 14,
    'RoadRaceSlopeBaseGTCurve2Out': 15,
    'RoadRaceSlopeBaseGTCurve2OutMirror': 16,
    'RoadRaceCheckpointMirror': 17,
    'RoadRaceCheckpoint': 18,
    'RoadRaceStartMirror': 19,
    'RoadRaceStart': 20,
    'RoadRaceFinish': 21,
    'RoadRaceFinishMirror': 22,
    'RoadRaceMultilapMirror': 23,
    'RoadRaceMultilap': 24,
    'RoadRaceTurboMirror': 25,
    'RoadRaceTurbo': 26,
    'RoadRaceTurbo2Mirror': 27,
    'RoadRaceTurbo2': 28,
    'RoadRaceFWMirror': 29,
    'RoadRaceFW': 30,
    'RoadRaceSlopeBase1': 31,
    'RoadRaceSlopeBase1Mirror': 32,
    'RoadRaceSlopeBase2': 33,
    'RoadRaceSlopeBase2Mirror': 34,
    'RoadRaceSlopeStaight': 35,
    'RoadRaceSlopeStaightMirror': 36,
    'RoadRaceSwitchSlopeStraight1x1Mirror': 37,
    'RoadRaceSwitchSlopeStraight1x1': 38,
    'RoadRaceBiSlopeStart': 39,
    'RoadRaceBiSlopeStartMirror': 40,
    'RoadRaceBiSlopeEnd': 41,
    'RoadRaceBiSlopeEndMirror': 42,
    'RoadRaceSlopeCheckpointUp': 43,
    'RoadRaceSlopeCheckpointUpMirror': 44,
    'RoadRaceSlopeCheckpointDown': 45,
    'RoadRaceSlopeCheckpointDownMirror': 46,
    'RoadRaceSlopeTurboUp': 47,
    'RoadRaceSlopeTurboUpMirror': 48,
    'RoadRaceSlopeTurboDown': 49,
    'RoadRaceSlopeTurboDownMirror': 50,
    'RoadRaceSlopeTurbo2Up': 51,
    'RoadRaceSlopeTurbo2UpMirror': 52,
    'RoadRaceSlopeTurbo2Down': 53,
    'RoadRaceSlopeTurbo2DownMirror': 54,
    'RoadRaceSlopeFW': 55,
    'RoadRaceSlopeFWMirror': 56,
    'RoadRaceTiltStraight': 57,
    'RoadRaceTiltCorner': 58,
    'RoadRaceTiltGTCurve2In': 59,
    'RoadRaceTiltCornerInUp': 60,
    'RoadRaceTiltCornerInUpMirror': 61,
    'RoadRaceTiltCornerOutDown': 62,
    'RoadRaceTiltCornerOutDownMirror': 63,
    'RoadRaceTiltGTCurve2InUp': 64,
    'RoadRaceTiltGTCurve2InUpMirror': 65,
    'RoadRaceTiltGTCurve2OutDown': 66,
    'RoadRaceTiltGTCurve2OutDownMirror': 67,
    'RoadRaceTiltCheckpoint': 68,
    'RoadRaceTiltCheckpointMirror': 69,
    'RoadRaceTiltTurbo': 70,
    'RoadRaceTiltTurboMirror': 71,
    'RoadRaceTiltTurbo2': 72,
    'RoadRaceTiltTurbo2Mirror': 73,
    'RoadRaceTiltFW': 74,
    'RoadRaceTiltToRoadRacex2': 75,
    'RoadRaceTiltToRoadRacex2Mirror': 76,
    'RoadRaceTiltToRoadRaceDiagIn': 77,
    'RoadRaceTiltToRoadRaceDiagInMirror': 78,
    'RoadRaceTiltToRoadRaceDiagOut': 79,
    'RoadRaceTiltToRoadRaceDiagOutMirror': 80,
    'RoadRaceToRoadMainA': 81,
    'RoadRaceToRoadMainMirrorA': 82,
    'RoadRaceToArena': 83,
    'RoadRaceToArenaMirror': 84,
    'ArenaSimpleBase': 85,
    'ArenaSimpleBaseNoPillar': 86,
    'ArenaBaseOut': 87,
    'ArenaSimpleDiag': 88,
    'ArenaSlopeBase': 89,
    'ArenaSlopeBase2': 90,
    'ArenaSlopeStart': 91,
    'ArenaSlopeStraight': 92,
    'ArenaSlopeEnd': 93,
    'ArenaBiSlopeStart': 94,
    'ArenaBiSlopeStraight': 95,
    'ArenaBiSlopeEnd': 96,
    'ArenaCheckpointBase': 97,
    'ArenaCheckpointSlopeStraightUp': 98,
    'ArenaCheckpointSlopeStraightDown': 99,
    'ArenaCheckpointSlopeStraightLeftAir': 100,
    'ArenaCheckpointSlopeStraightRightAir': 101,
    'OffRoadCheckpoint': 102,
    'ArenaStart': 103,
    'ArenaFinish': 104,
    'ArenaMultilap': 105,
    'ArenaTurbo1': 106,
    'ArenaTurbo1SlopeUp': 107,
    'ArenaTurbo1SlopeDown': 108,
    'ArenaTurbo1SlopeLeft': 109,
    'ArenaTurbo1SlopeRight': 110,
    'ArenaTurbo2': 111,
    'ArenaTurbo2SlopeUp': 112,
    'ArenaTurbo2SlopeDown': 113,
    'ArenaTurbo2SlopeLeft': 114,
    'ArenaTurbo2SlopeRight': 115,
    'ArenaFreeWheeling': 116,
    'ArenaFWSlopeUD': 117,
    'ArenaFWSlopeLR': 118,
    'ArenaBridgeBase': 119,
    'ArenaBridgeBaseNoPillar': 120,
    'ArenaSimpleHalfBase': 121,
    'ArenaSimpleHalfBaseCornerIn': 122,
    'ArenaSimpleHalfBaseCornerOut': 123,
    'ArenaRamp': 124,
    'ArenaFloodLightBaseU': 125,
    'ArenaFloodLightBaseCeilingA': 126,
    'ArenaScreenBase2x1': 127,
    'DecoLightArenaStraightLeft': 128,
    'DecoLightArenaStraightMed': 129,
    'DecoLightArenaStraightRight': 130,
    'ArenaTransitionFromRoadMain2': 131,
    'ArenaTransitionBridgeToRoadMain': 132,
    'ArenaTransitionToRoadMain': 133,
    'ArenaWallRideStraight': 134,
    'ArenaWallRideCorner': 135,
    'ArenaWallRideStraightGTCurve2': 136,
    'ArenaWallRideScreen2x1': 137,
    'ArenaWallRideCheckpoint_H': 138,
    'ArenaWallRideCheckpoint_V': 139,
    'ArenaWallRideStartStraight': 140,
    'ArenaWallRideStartCorner': 141,
    'ArenaWallRideStartGTCurve2': 142,
    'ArenaWallRideLoopStart': 143,
    'ArenaWallRideLoopEnd': 144,
    'ArenaWallRideTopStraight': 145,
    'ArenaWallRideTopCorner': 146,
    'ArenaWallRideTopGTCurve2': 147,
    'ArenaWallRideBorderStraight': 148,
    'ArenaWallRideBorderCorner': 149,
    'ArenaWallRideBorderGTCurve2': 150,
    'ArenaBorderStraightTower': 151,
    'ArenaBorderStraightA': 152,
    'ArenaBorderCornerTowerA': 153,
    'ArenaBorderCorner': 154,
    'ArenaBorderStraightPillarBank': 155,
    'ArenaBorderStraightPillarBankToArena': 156,
    'ArenaBorderCornerPillarBank': 157,
    'ArenaPasserelleStraight': 158,
    'ArenaRoadMainAndBank': 159,
    'ArenaTransitionFromRoadMainAndBankTunnel': 160,
    'ArenaTransitionFromRoadMainAndBankTunnelMirror': 161,
    'ArenaTransitionFromRoadMaindAndBank': 162,
    'ArenaTransitionFromRoadMaindAndBankMirror': 163,
    'ArenaHallwayStraight': 164,
    'ArenaHallway2WaysStraight': 165,
    'ArenaHallwayCorner': 166,
    'ArenaWallRideSlopeStraight': 167,
    'ArenaHallwaySlopeStraight': 168,
    'ArenaWallRideSlopeHallWay': 169,
    'ArenaLoopStraight1': 170,
    'ArenaLoopCornerOut': 171,
    'ArenaLoopStraightSlope': 172,
    'ArenaLoopStraightSlopeMirror': 173,
    'ArenaLoopStraightSlope2': 174,
    'ArenaLoopStraightSlope2Mirror': 175,
    'ArenaTunnelStraight': 176,
    'ArenaTunnelStraightOpen1Side': 177,
    'ArenaTunnelStraightOpen2Sides': 178,
    'ArenaTunnelStraightToOpen': 179,
    'ArenaTunnelCheckpoint': 180,
    'ArenaTunnelStraightOpen2SidesCheckpoint': 181,
    'ArenaTunnelCorner': 182,
    'ArenaTunnelGTCurve2': 183,
    'ArenaTunnelGTCurve3': 184,
    'ArenaTunnelGTDiag2': 185,
    'ArenaTunnelGTDiag2Mirror': 186,
    'ArenaTunnelGTDiag3': 187,
    'ArenaTunnelGTDiag3Mirror': 188,
    'ArenaTunnelSlopeBase': 189,
    'ArenaTunnelSlopeStart': 190,
    'ArenaTunnelSlopeStraight2': 191,
    'ArenaTunnelSlopeEnd': 192,
    'ArenaTunnelBiSlopeStart': 193,
    'ArenaTunnelSlopeStraight': 194,
    'ArenaTunnelBiSlopeEnd': 195,
    'ArenaTunnelSlopeBase2': 196,
    'ArenaTunnelSlopeBaseHole': 197,
    'ArenaTunnelSlopeBaseHole2': 198,
    'ArenaTunnelBanktoRoad': 199,
    'ArenaTunnelToArena': 200,
    'ArenaTunnelTransitionToBankLeft': 201,
    'ArenaTunnelTransitionToBankRight': 202,
    'ArenaTubeStraightA': 203,
    'ArenaTubeStraightB': 204,
    'TubeStraightScreenDown': 205,
    'TubeStraightScreenSide': 206,
    'TubeStraightScreenUp': 207,
    'TubeCrossAir': 208,
    'TubeCrossV': 209,
    'VStraight1Screen': 210,
    'TubeVStraight1AirA': 211,
    'TubeVStraight2AirA': 212,
    'TubeVStraight4AirA': 213,
    'ArenaTubeCorner': 214,
    'ArenaTubeGTCurve2': 215,
    'TubeGTCurve3': 216,
    'ToTubeVCornerAirAUp': 217,
    'TubeGTCurve2Up': 218,
    'ToTubeVCornerAirADown': 219,
    'TubeGTCurve2Down': 220,
    'TubeLoopRight': 221,
    'TubeLoopLeft': 222,
    'TubeCheckpoint': 223,
    'TubeCheckpointVAir': 224,
    'TubeFinish': 225,
    'TubeFreeWheeling': 226,
    'TubeToRoadMain': 227,
    'TubeToGround': 228,
    'TubeToArenaCenter': 229,
    'TubeStraightUnderground': 230,
    'TubeStraightScreenDownUnderground': 231,
    'TubeStraightScreenSideUnderground': 232,
    'TubeStraightScreenUpUnderground': 233,
    'TubeCrossUnderground': 234,
    'TubeCrossVUnderground': 235,
    'TubeVStraight1ScreenUnderground': 236,
    'TubeVStraight1Underground': 237,
    'TubeVStraight2Underground': 238,
    'ToTubeVStraight4UndergroundA': 239,
    'TubeCornerUnderground': 240,
    'ArenaTubeGTCurve2Underground': 241,
    'TubeGTCurve3Underground': 242,
    'TubeCornerUndergroundUp': 243,
    'ArenaTubeGTCurve2UndergroundUp': 244,
    'TubeCornerUndergroundDown': 245,
    'TubeGTCurve2UndergroundDown': 246,
    'TubeCheckpointUnderground': 247,
    'TubeCheckpointVUnderground': 248,
    'TubeFinishUnderground': 249,
    'TubeFreeWheelingUnderground': 250,
    'TubeLoopRightUnderground': 251,
    'TubeLoopLeftUnderground': 252,
    'TubeUndergroundDirtHill2': 253,
    'TubeUndergroundEntryDirtCliff4Low': 254,
    'TubeUndergroundEntryDirtCliff8Low': 255,
    'TubeUndergroundEntryDirtCliff8High': 256,
    'TubeUndergroundEntryWaterCliff12Low': 257,
    'TubeUndergroundEntryWaterCliff12Med': 258,
    'TubeUndergroundEntryWaterCliff12High': 259,
    'TubeUndergroundToRoadCave': 260,
    'RoadMain': 261,
    'RoadMainTreePineBigStraight': 262,
    'RoadMainJumpSmall': 263,
    'RoadMainStraightUp': 264,
    'RoadMainStraightDown': 265,
    'RoadMainSupport': 266,
    'RoadMainGTCurve2': 267,
    'RoadMainSlopeGTCurve2': 268,
    'RoadMainSlopeGTCurve2Mirror': 269,
    'RoadMainSlopeBase': 270,
    'RoadMainSlopeBase3': 271,
    'RoadMainBiSlopeStart': 272,
    'RoadMainSlopeStraight': 273,
    'RoadMainBiSlopeEnd': 274,
    'RoadMainGTDiag2X2A': 275,
    'RoadMainGTDiag2X2MirrorA': 276,
    'RoadMainGTDiag4X2': 277,
    'RoadMainGTDiag4X2Mirror': 278,
    'RoadMainGTDiag2X2Down': 279,
    'RoadMainGTDiag2X2DownMirror': 280,
    'RoadMainBridgeDirtHill2': 281,
    'RoadMainCheckpoint': 282,
    'RoadMainCheckpointUp': 283,
    'RoadMainBridgeDirtCliff4': 284,
    'RoadMainBridgeDirtCliff8': 285,
    'RoadMainCheckpointDown': 286,
    'RoadMainBridgeWaterCliff12': 287,
    'RoadMainStart': 288,
    'RoadMainFinish': 289,
    'RoadMainMultilap': 290,
    'RoadMainTurbo': 291,
    'RoadMainTurbo2': 292,
    'RoadMainFW': 293,
    'RoadHill2Straight': 294,
    'RoadHill2CornerIn': 295,
    'RoadHill2CornerOut': 296,
    'RoadHill2GTCurve2In': 297,
    'RoadHill2GTCurve2Out': 298,
    'RoadHill2CheckpointLeft': 299,
    'RoadHill2CheckpointRight': 300,
    'RoadMain2ToHill2Corner': 301,
    'RoadMain2ToHill2CornerMirror': 302,
    'Cliff12RoadHighStraight': 303,
    'Cliff12RoadHighBarrageLeft': 304,
    'Cliff12RoadHighBarrageRight': 305,
    'Cliff12RoadHighObstaclesRocks3x1': 306,
    'Cliff12RoadHighLoopHighLeft': 307,
    'Cliff12RoadHighLoopHighRight': 308,
    'Cliff12RoadHighBase1': 309,
    'RoadRaceBridgeDirtHill2': 310,
    'RoadRaceBridgeDirtHill2Mirror': 311,
    'Cliff12RoadHighCorner': 312,
    'RoadRaceBridgeDirtCliff4': 313,
    'RoadRaceBridgeDirtCliff4Mirror': 314,
    'RoadRaceBridgeDirtCliff8': 315,
    'Cliff12RoadHighGTCurve2In': 316,
    'RoadRaceBridgeDirtCliff8Mirror': 317,
    'Cliff12RoadHighGTCurve2Out': 318,
    'RoadRaceBridgeWaterCliff12': 319,
    'Cliff12RoadHighGTDiag3': 320,
    'RoadRaceBridgeWaterCliff12Mirror': 321,
    'Cliff12RoadHighGTDiag3Mirror': 322,
    'Cliff12RoadHighCheckpointLeft': 323,
    'Cliff12RoadHighCheckpointRight': 324,
    'Cliff12RoadHighTurboLeft': 325,
    'Cliff12RoadHighTurbo': 326,
    'Cliff12RoadHighTurboLeft2': 327,
    'Cliff12RoadHighTurbo2': 328,
    'Cliff12RoadHighFW': 329,
    'Cliff12RoadHighToRoadMainA': 330,
    'Cliff12RoadHighToRoadMainA_mirror': 331,
    'Cliff12RoadHighToRoadAirA': 332,
    'Cliff12RoadHighToRoadAirA_mirror': 333,
    'Cliff12RoadMedStraight': 334,
    'Cliff12RoadMedBarrageLeft': 335,
    'Cliff12RoadMedBarrageRight': 336,
    'Cliff12RoadMedObstaclesRocks3x1': 337,
    'Cliff12RoadMedLoopLeft': 338,
    'Cliff12RoadMedLoopRight': 339,
    'Cliff12RoadMedBase1': 340,
    'Cliff12RoadMedCorner': 341,
    'Cliff12RoadMedGTCurve2IN': 342,
    'Cliff12RoadMedGTCurve2OUT': 343,
    'Cliff12RoadMedGTDiag3': 344,
    'Cliff12RoadMedGTDiag3Mirror': 345,
    'Cliff12RoadMedCheckpointLeft': 346,
    'Cliff12RoadMedCheckpointRight': 347,
    'Cliff12RoadMedTurboLeft': 348,
    'Cliff12RoadMedTurbo': 349,
    'Cliff12RoadMedTurbo2Left': 350,
    'Cliff12RoadMedTurbo2': 351,
    'Cliff12RoadMedFW': 352,
    'Cliff12RoadHighToMedSlope3A': 353,
    'Cliff12RoadHighToMedSlope3A_Mirror': 354,
    'Cliff12RoadMedToRoadAir': 355,
    'Cliff12RoadMedToRoadAir_mirror': 356,
    'RoadCaveStraight': 357,
    'RoadCaveObstaclesYShaped2x1': 358,
    'RoadCaveCorner': 359,
    'RoadCaveGTCurve2': 360,
    'RoadCaveGTCurve3': 361,
    'RoadCaveGTDiag2': 362,
    'RoadCaveGTDiag2Mirror': 363,
    'RoadCaveGTDiag3': 364,
    'RoadCaveGTDiag3Mirror': 365,
    'RoadCaveSlopeBase': 366,
    'RoadCaveSlopeBase4': 367,
    'RoadCaveSlopeBaseGtCurve3': 368,
    'RoadCaveSlopeBaseGtCurve3Mirror': 369,
    'RoadCaveSlope2Base3x1': 370,
    'RoadCaveSlope2Start': 371,
    'RoadCaveSlope2Start2x2x1': 372,
    'RoadCaveSlope2Straight': 373,
    'RoadCaveSlope2End': 374,
    'RoadCaveSlope2End2x2x1': 375,
    'RoadCaveSlope2UTop': 376,
    'RoadCaveSlope2UBottom': 377,
    'RoadCaveSlope2Base3x1GT3Left': 378,
    'RoadCaveSlope2Base3x1GT3Right': 379,
    'RoadCaveCheckpoint': 380,
    'RoadCaveCheckpointSlope2Up': 381,
    'RoadCaveCheckpointSlope2Down': 382,
    'RoadCaveFinish': 383,
    'RoadCaveMultiLap': 384,
    'RoadCaveTurbo': 385,
    'RoadCaveTurboUp': 386,
    'RoadCaveTurboDown': 387,
    'RoadCaveTurbo2': 388,
    'RoadCaveTurbo2Up': 389,
    'RoadCaveTurbo2Down': 390,
    'RoadCaveFW': 391,
    'RoadCaveFWSlope2': 392,
    'RoadEntrySlopeStraight': 393,
    'RoadCaveEntryDirtHill2': 394,
    'RoadCaveEntryDirtCliff8Low': 395,
    'RoadCaveEntryDirtCliff4Low': 396,
    'RoadCaveEntryDirtCliff4Low2': 397,
    'RoadCaveEntryDirtCliff8Low2': 398,
    'RoadCaveEntryDirtCliff8High': 399,
    'RoadCaveEntryDirtCliff12Low': 400,
    'RoadCaveEntryDirtCliff12Med': 401,
    'RoadCaveEntryDirtCliff12High': 402,
    'RoadCaveDoubleStraight': 403,
    'RoadCaveDoubleObstacleRocks': 404,
    'RoadCaveDoubleObstacleJumpYShapedToGround': 405,
    'RoadCaveDoubleCross': 406,
    'RoadCaveDoubleObstacleJumpRoadCave': 407,
    'RoadCaveDoubleDeadend': 408,
    'RoadCaveDoubleCorner': 409,
    'RoadCaveDoubleGTCurve2': 410,
    'RoadCaveDoubleGTCurve3': 411,
    'RoadCaveDoubleDiag2x2': 412,
    'RoadCaveDoubleDiag2x2Mirror': 413,
    'RoadCaveDoubleDiag3x2': 414,
    'RoadCaveDoubleDiag3x2Mirror': 415,
    'RoadCaveDoubleSwitch2x1': 416,
    'RoadCaveDoubleSwitch3x1': 417,
    'RoadCaveDoubleSwitch3x1Mirror': 418,
    'RoadCaveDoubleSwitchDiag3x2': 419,
    'RoadCaveDoubleSwitchDiag3x2Mirror': 420,
    'RoadCaveDoubleCheckpoint': 421,
    'RoadCaveDoubleCheckpointCenter': 422,
    'RoadCaveDoubleCheckpointLeft': 423,
    'RoadCaveDoubleCheckpointRight': 424,
    'RoadCaveDoubleFinish': 425,
    'RoadCaveDoubleSlope2Base': 426,
    'RoadCaveDoubleSlope2Base2x1': 427,
    'RoadCaveDoubleSlope2Start': 428,
    'RoadCaveDoubleSlope2Straight': 429,
    'RoadCaveDoubleSlope2End': 430,
    'RoadCaveDoubleSlope2UTop': 431,
    'RoadCaveDoubleSlope2UBottom': 432,
    'RoadCaveDoubleToRoadCaveStraight': 433,
    'RoadCaveDoubleToRoadCaveCorner': 434,
    'RoadCaveDoubleToRoadCaveCornerMirror': 435,
    'RoadCaveDoubleToRoadCaveTShaped': 436,
    'DecoTreesBig1X1': 437,
    'DecoTreesMedium1X1': 438,
    'DecoRocksMedium1x1': 439,
    'DecoCactusMedium1X1': 440,
    'DecoBushesMedium1X1': 441,
    'Sequoias': 442,
    'DecoTerrainHD': 443,
    'ScreenBig2x1': 444,
    'Screen2x1StraightA': 445,
    'Screen2x1SupportStraightA': 446,
    'Screen2x1SupportCornerA': 447,
    'Screen2x1Small': 448,
    'Screen2x1SmallCenter': 449,
    'DecoTechnicsAntenna': 450,
    'Podium': 451,
    'DecoReffineryX1': 452,
    'DecoReffineryX2': 453,
    'DecoReffineryX4': 454,
    'DecoPipelineStraight': 455,
    'DecoPipelineCorner': 456,
    'DecoPipelineGTCurve2': 457,
    'DecoPipelineTshape': 458,
    'DecoPipelineSlope': 459,
    'PipelineDiag2x3Mirror': 460,
    'PipelineDiag2x3': 461,
}


STADIUM_BLOCKS = {
    'StadiumGrassClip': 1,
    'StadiumDirt': 2,
    'StadiumDirtBorder': 3,
    'StadiumWater': 4,
    'StadiumPool': 5,
    'StadiumRoadMain': 6,
    'StadiumRoadMainGTCurve2': 7,
    'StadiumRoadMainGTCurve3': 8,
    'StadiumRoadMainGTCurve4': 9,
    'StadiumRoadMainGTCurve5': 10,
    'StadiumLoopLeft': 11,
    'StadiumLoopRight': 12,
    'StadiumRoadMainCheckpoint': 13,
    'StadiumRoadMainStartLine': 14,
    'StadiumRoadMainFinishLine': 15,
    'StadiumRoadMainStartFinishLine': 16,
    'StadiumRoadMainTurbo': 17,
    'StadiumRoadMainTurboRoulette': 18,
    'StadiumRoadMainFW': 19,
    'StadiumGrassCheckpoint': 20,
    'StadiumHolePillar': 21,
    'StadiumHolePillar2Front': 22,
    'StadiumHolePillar2Line': 23,
    'StadiumHolePillar3': 24,
    'StadiumHole': 25,
    'StadiumRoadStretch': 26,
    'StadiumBump1': 27,
    'StadiumRamp': 28,
    'StadiumRampLow': 29,
    'StadiumRoadMainGTDiag2x2': 30,
    'StadiumRoadMainGTDiag2x2Mirror': 31,
    'StadiumRoadMainGTDiag3x2': 32,
    'StadiumRoadMainGTDiag3x2Mirror': 33,
    'StadiumRoadMainGTDiag4x3': 34,
    'StadiumRoadMainGTDiag4x3Mirror': 35,
    'StadiumRoadMainYShapedDiag2': 36,
    'StadiumRoadMainYShapedCurve2': 37,
    'StadiumRoadMainSlopeBase2': 38,
    'StadiumRoadMainSlopeBase': 39,
    'StadiumRoadMainSlopeStraight': 40,
    'StadiumRoadMainBiSlopeStart': 41,
    'StadiumRoadMainBiSlopeEnd': 42,
    'StadiumRoadMainSlopeUTop': 43,
    'StadiumRoadMainSlopeUBottom': 44,
    'StadiumRoadMainCheckpointUp': 45,
    'StadiumRoadMainCheckpointDown': 46,
    'StadiumRoadMainTurboUp': 47,
    'StadiumRoadMainTurboDown': 48,
    'StadiumRoadMainTurboRouletteDown': 49,
    'StadiumRoadMainTurboRouletteUp': 50,
    'StadiumHolePillarSlope': 51,
    'StadiumHoleSlope': 52,
    'StadiumRoadStretchSlope': 53,
    'StadiumBump1Slope': 54,
    'StadiumRoadMainGTDiag2x2Slope': 55,
    'StadiumRoadMainGTDiag2x2SlopeMirror': 56,
    'StadiumRoadMainGTDiag3x2Slope': 57,
    'StadiumRoadMainGTDiag3x2SlopeMirror': 58,
    'StadiumRoadMainGTDiag4x3Slope': 59,
    'StadiumRoadMainYShapedDiag2SlopeUp': 60,
    'StadiumRoadMainYShapedDiag2SlopeDown': 61,
    'StadiumRoadMainGTDiag4x3SlopeMirror': 62,
    'StadiumRoadTiltStraight': 63,
    'StadiumRoadTiltTransition1Left': 64,
    'StadiumRoadTiltTransition1Right': 65,
    'StadiumRoadTiltTransition2Left': 66,
    'StadiumRoadTiltTransition2Right': 67,
    'StadiumRoadTiltTransition2CurveLeft': 68,
    'StadiumRoadTiltTransition2CurveRight': 69,
    'StadiumRoadTiltTransition2DiagLeft': 70,
    'StadiumRoadTiltTransition2DiagRight': 71,
    'StadiumRoadMainCheckpointLeft': 72,
    'StadiumRoadMainCheckpointRight': 73,
    'StadiumRoadMainTurboLeft': 74,
    'StadiumRoadMainTurboRight': 75,
    'StadiumRoadMainTurboRouletteLeft': 76,
    'StadiumRoadMainTurboRouletteRight': 77,
    'StadiumRoadTiltCorner': 78,
    'StadiumRoadTiltGTCurve2': 79,
    'StadiumRoadTiltGTCurve3': 80,
    'StadiumRoadTiltGTCurve4': 81,
    'StadiumRoadMainGTDiag2x2Tilt': 82,
    'StadiumRoadMainGTDiag2x2TiltMirror': 83,
    'StadiumRoadMainGTDiag3x2Tilt': 84,
    'StadiumRoadMainGTDiag3x2TiltMirror': 85,
    'StadiumRoadTiltCornerDownLeft': 86,
    'StadiumRoadTiltCornerDownRight': 87,
    'StadiumRoadTiltCornerUpLeft': 88,
    'StadiumRoadTiltCornerUpRight': 89,
    'StadiumRoadTiltGTCurve2DownLeft': 90,
    'StadiumRoadTiltGTCurve2DownRight': 91,
    'StadiumRoadTiltGTCurve2UpLeft': 92,
    'StadiumRoadTiltGTCurve2UpRight': 93,
    'StadiumHolePillarTilt': 94,
    'StadiumHoleTilt': 95,
    'StadiumStretchTilt': 96,
    'StadiumBump1Tilt': 97,
    'StadiumPlatformLoopStart': 98,
    'StadiumPlatformLoopEnd': 99,
    'StadiumPlatformWallBorder': 100,
    'StadiumPlatformWall1': 101,
    'StadiumPlatformWall2': 102,
    'StadiumPlatformWall4': 103,
    'StadiumPlatformWallPub2': 104,
    'StadiumPlatformRoad': 105,
    'StadiumPlatformSlope2Straight': 106,
    'StadiumPlatformBiSlope2Start': 107,
    'StadiumPlatformBiSlope2End': 108,
    'StadiumPlatformTurbo': 109,
    'StadiumPlatformTurboDown': 110,
    'StadiumPlatformTurboLeft': 111,
    'StadiumPlatformTurboRight': 112,
    'StadiumPlatformTurboUp': 113,
    'StadiumCheckpointRingV': 114,
    'StadiumCheckpointRingHRoad': 115,
    'StadiumPlatformCheckpoint': 116,
    'StadiumPlatformCheckpointDown': 117,
    'StadiumPlatformCheckpointLeft': 118,
    'StadiumPlatformCheckpointRight': 119,
    'StadiumPlatformCheckpointUp': 120,
    'StadiumPlatformGTCurve2Wall1': 121,
    'StadiumPlatformGTCurve2Wall2': 122,
    'StadiumPlatformGTCurve2Wall4': 123,
    'StadiumPlatformGTCurve3Wall1': 124,
    'StadiumPlatformGTCurve3Wall2': 125,
    'StadiumPlatformGTCurve3Wall4': 126,
    'StadiumPlatformToRoad': 127,
    'StadiumPlatformToRoad2': 128,
    'StadiumPlatformToRoad2Mirror': 129,
    'StadiumPlatformToRoadMain': 130,
    'StadiumCircuitBorderStraight': 131,
    'StadiumCircuitBorderDiagIn': 132,
    'StadiumCircuitBorderDiagOut': 133,
    'StadiumCircuitBorderCornerIn': 134,
    'StadiumCircuitBorderCornerOut': 135,
    'StadiumCircuitBase': 136,
    'StadiumCircuitSlopeStart': 137,
    'StadiumCircuitSlopeStraight': 138,
    'StadiumCircuitSlopeEnd': 139,
    'StadiumCircuitTurbo': 140,
    'StadiumCircuitTurboLeft': 141,
    'StadiumCircuitTurboUp': 142,
    'StadiumCircuitTurboRight': 143,
    'StadiumCircuitTurboDown': 144,
    'StadiumCircuitBorderGTCurve2In': 145,
    'StadiumCircuitBorderGTCurve3In': 146,
    'StadiumCircuitBorderGTCurve4In': 147,
    'StadiumCircuitBorderGTCurve2Out': 148,
    'StadiumCircuitBorderGTCurve3Out': 149,
    'StadiumCircuitBorderGTCurve4Out': 150,
    'StadiumCircuitBorderSlopeStartLeft': 151,
    'StadiumCircuitBorderSlopeStartRight': 152,
    'StadiumCircuitBorderSlopeStraightLeft': 153,
    'StadiumCircuitBorderSlopeStraightRight': 154,
    'StadiumCircuitBorderSlopeStraightBottom': 155,
    'StadiumCircuitBorderSlopeStraightTop': 156,
    'StadiumCircuitBorderSlopeEndLeft': 157,
    'StadiumCircuitBorderSlopeEndRight': 158,
    'StadiumCircuitBorderSlopeCornerInLeftBottom': 159,
    'StadiumCircuitBorderSlopeCornerInLeftTop': 160,
    'StadiumCircuitBorderSlopeCornerInRightTop': 161,
    'StadiumCircuitBorderSlopeCornerInRightBottom': 162,
    'StadiumCircuitSlopeCornerInLeftBottom': 163,
    'StadiumCircuitBorderSlopeCornerOutLeftTop': 164,
    'StadiumCircuitBorderSlopeCornerOutRightTop': 165,
    'StadiumCircuitBorderSlopeCornerOutRightBottom': 166,
    'StadiumCircuitBorderSlopeGTCurve2InLeftBottom': 167,
    'StadiumCircuitBorderSlopeGTCurve2InLeftTop': 168,
    'StadiumCircuitBorderSlopeGTCurve2InRightTop': 169,
    'StadiumCircuitBorderSlopeGTCurve2InRightBottom': 170,
    'StadiumCircuitBorderSlopeGTCurve2OutLeftBottom': 171,
    'StadiumCircuitBorderSlopeGTCurve2OutLeftTop': 172,
    'StadiumCircuitBorderSlopeGTCurve2OutRightTop': 173,
    'StadiumCircuitBorderSlopeGTCurve2OutRightBottom': 174,
    'StadiumCircuitBorderSlopeDiagInLeftBottom': 175,
    'StadiumCircuitBorderSlopeDiagInLeftTop': 176,
    'StadiumCircuitBorderSlopeDiagInRightTop': 177,
    'StadiumCircuitBorderSlopeDiagInRightBottom': 178,
    'StadiumCircuitBorderSlopeDiagOutLeftBottom': 179,
    'StadiumCircuitBorderSlopeDiagOutLeftTop': 180,
    'StadiumCircuitBorderSlopeDiagOutRightTop': 181,
    'StadiumCircuitBorderSlopeDiagOutRightBottom': 182,
    'StadiumCircuitLoopStart': 183,
    'StadiumCircuitRampBig': 184,
    'StadiumCircuitRampSmall1x05': 185,
    'StadiumCircuitRampSmall1x1': 186,
    'StadiumCircuitHole': 187,
    'StadiumCircuitBumpUp': 188,
    'StadiumCircuitBumpUp2': 189,
    'StadiumCircuitPillar1': 190,
    'StadiumCircuitPillar2': 191,
    'StadiumCircuitPillar3': 192,
    'StadiumCircuitPillar5': 193,
    'StadiumCircuitPillar6': 194,
    'StadiumCircuitPillar7': 195,
    'StadiumRoadDirtFenceStraight': 196,
    'StadiumRoadDirtFenceCorner': 197,
    'StadiumRoadDirt': 198,
    'StadiumRoadDirtTurbo': 199,
    'StadiumRoadDirtCheckpoint': 200,
    'StadiumRoadDirtGTCurve2': 201,
    'StadiumRoadDirtGTCurve3': 202,
    'StadiumRoadDirtDiagonaleRight': 203,
    'StadiumRoadDirtDiagonaleLeft': 204,
    'StadiumRoadDirtHighFenceStraight': 205,
    'StadiumRoadDirtHighFenceCorner': 206,
    'StadiumRoadDirtHigh': 207,
    'StadiumRoadDirtHighCheckpoint': 208,
    'StadiumRoadDirtHighGTCurve3': 209,
    'BStadiumRoadDirtToRoadDirtHigh': 210,
    'StadiumRoadDirtToRoadDirtHigh2': 211,
    'StadiumRoadDirtToRoadDirtHighBridge': 212,
    'StadiumRoadDirtToRoadDirtHighCross': 213,
    'StadiumRoadDirtToRoadDirtHigh': 214,
    'StadiumRoadDirtToRoad': 215,
    'StadiumRoadDirtHillSlope2': 216,
    'StadiumRoadDirtHillSlope': 217,
    'StadiumRoadDirtHillWave': 218,
    'StadiumRoadDirtHillSlopeGT2Bis': 219,
    'StadiumRoadDirtHillSlopeGT2': 220,
    'StadiumRoadDirtWave_x2': 221,
    'StadiumRoadDirtWave_x1': 222,
    'StadiumRoadDirtToRoadGrass': 223,
    'StadiumDirtDeadendDoor': 224,
    'StadiumDirtDeadendSpot': 225,
    'StadiumDirtDeadendTent1': 226,
    'StadiumRoadDirtHillTiltStraight': 227,
    'StadiumRoadDirtHillTiltCornerIn': 228,
    'StadiumRoadDirtHillTiltCornerOut': 229,
    'StadiumRoadDirtHillTiltCornerLeft': 230,
    'StadiumRoadDirtHillTiltCornerRight': 231,
    'StadiumRoadDirtHillTiltToRoadLeft': 232,
    'StadiumRoadDirtHillTiltToRoadRight': 233,
    'StadiumSculptBridgeSuspendSlope2': 234,
    'StadiumSculptBridgeSuspendSlope2Mirror': 235,
    'StadiumSculptBridgeSuspendSlopeEnd': 236,
    'StadiumSculptBridgeSuspendSlopeEndMirror': 237,
    'StadiumSculptA': 238,
    'StadiumSculptB': 239,
    'StadiumSculptC': 240,
    'StadiumSculptArchRingEnd': 241,
    'StadiumSculptArchRingEndMirror': 242,
    'StadiumSculptArchRingStart': 243,
    'StadiumSculptArchRingStartMirror': 244,
    'StadiumSculptBridgePillar': 245,
    'StadiumSculptBridgePillarMirror': 246,
    'StadiumSculptBridgeSlopeStart': 247,
    'StadiumSculptBridgeSlopeStartMirror': 248,
    'StadiumSculptBridgeSlopeEnd': 249,
    'StadiumSculptBridgeSlopeEndMirror': 250,
    'StadiumSculptBridgeStraight': 251,
    'StadiumSculptBridgeStraightSmall': 252,
    'StadiumTubePillarCap': 253,
    'StadiumTubeV1': 254,
    'StadiumTubeV4': 255,
    'StadiumTubeV8': 256,
    'StadiumTube': 257,
    'StadiumTubePillar': 258,
    'StadiumTubePillarBranch': 259,
    'StadiumTubeRoad': 260,
    'StadiumTubeRoadDown': 261,
    'StadiumTubeRoadUp': 262,
    'StadiumTubeRoadCross': 263,
    'StadiumControlRoadGlass': 264,
    'StadiumControlRoadPub': 265,
    'StadiumControlRoadCamera': 266,
    'StadiumControlCameraPub': 267,
    'StadiumControlLight': 268,
    'StadiumControlLightBase': 269,
    'StadiumTubeRoadLightSystem': 270,
    'StadiumTubeRoadSoundSystem': 271,
    'StadiumFabricStraight1x1': 272,
    'StadiumFabricPillarAir': 273,
    'StadiumFabricCornerIn': 274,
    'StadiumFabricPillarCornerInAir': 275,
    'StadiumFabricCornerOut': 276,
    'StadiumFabricPillarCornerOut': 277,
    'StadiumFabricCross1x1': 278,
    'StadiumFabricCross3x3': 279,
    'StadiumFabricCross3x3Screen': 280,
    'StadiumFabricPillarAirScreen': 281,
    'StadiumInflatableSupport': 282,
    'StadiumInflatablePillar': 283,
    'StadiumInflatableTube': 284,
    'StadiumInflatableAdvert': 285,
    'StadiumAirship': 286,
    'StadiumInflatableCastle': 287,
    'StadiumInflatableCastleDoor': 288,
    'StadiumInflatableCastleBig': 289,
    'StadiumInflatableCactus': 290,
    'StadiumInflatableSnowTree': 291,
    'StadiumInflatablePalmTree': 292,

    # TM2 Stadium specifc blocks
    'StadiumRoadMainBirds': 293,
    'StadiumGrassBirds': 294,
    'StadiumRoadMainInBeam': 295,
    'StadiumRoadMainXBeam': 296,
    'StadiumRoadMainFWSlope': 297,
    'StadiumRoadMainInBeamSlope': 298,
    'StadiumRoadMainXBeamSlope': 299,
    'StadiumRoadMainTiltStraightBirds': 300,
    'StadiumRoadMainFWTilt': 301,
    'StadiumRoadMainInBeamTilt': 302,
    'StadiumRoadMainXBeamTilt': 303,
    'StadiumRoadMainSlopeBase1x2': 304,
    'StadiumBump1InGround': 305,
    'StadiumPlatformWallCheckpointH': 306,
    'StadiumPlatformWallCheckpointV': 307,
    'StadiumPlatformWall1Birds': 308,
    'StadiumPlatformFW': 309,
    'StadiumPlatformFWSlope2': 310,
    'StadiumPlatformMultilap': 311,
    'StadiumPlatformGridCheckpoint': 312,
    'StadiumPlatformGridFW': 313,
    'StadiumPlatformGridStraight': 314,
    'StadiumPlatformGridTurbo': 315,
    'StadiumPlatformGridStraightBirds': 316,
    'StadiumCheckpointRing2x1V': 317,
    'StadiumCheckpointRing2x1H': 318,
    'StadiumCircuitToRoadMain': 319,
    'StadiumCircuitFreeWheeling': 320,
    'StadiumCircuitFreeWheelingSlope2Air': 321,
    'StadiumRoadDirtGTCurve4': 322,
    'StadiumRoadDirtStartFinishLine': 323,
    'StadiumRoadDirtMultiLap': 324,
    'StadiumRoadDirtStraightBirds': 325,
    'StadiumRoadDirtHighMultiLap': 326,
    'StadiumRoadDirtHighDTCurve4': 327,
    'StadiumRoadDirtHighFenceStraightBirds': 328,
    'StadiumTrenchStraight': 329,
    'StadiumTrenchStraightTunnel': 330,
    'StadiumTrenchCorner': 331,
    'StadiumTrenchCross': 332,
    'StadiumTrenchCrossTunnel': 333,
    'StadiumTrenchGTCurve2': 334,
    'StadiumTrenchGTCurve3': 335,
    'StadiumTrenchGTCurve3Tunnel': 336,
    'StadiumTrenchGTCurve2Tunnel': 337,
    'StadiumTrenchCheckpoint': 338,
    'StadiumTrenchMultiLap': 339,
    'StadiumTrenchFreeWheeling': 340,
    'StadiumTrenchStraightScreen': 341,
    'StadiumTrenchInPillar': 342,
    'StadiumTrenchToRoadMain': 343,
    'StadiumTrenchToRoadMainBiSlopeStart': 344,
    'StadiumTrenchXRoadMain': 345,
    'StadiumTrenchToRoadDirt': 346,
    'StadiumTrenchToRoadDirtX2': 347,
    'StadiumTrenchToPlatformBiSlopeStart': 348,
    'StadiumTrenchToLoopStart': 349,
    'StadiumPodium': 350,
    'StadiumDecoTowerCore6': 351,
    'StadiumDecoTowerCore': 352,
    'StadiumDecoTowerTowerBeam': 353,
    'StadiumDecoTowerTowerA': 354,
    'StadiumDecoTowerTowerB': 355,
    'StadiumDecoTowerTowerD': 356,
    'StadiumDecoTowerTowerC': 357,

    # TODO: sort this out
    'StadiumRoadDirtHighToRoad': 358,
    'StadiumDirtHill': 359,
    'StadiumPlatformBiSlope2StartSmall': 360,
    'StadiumWaterClip': 361,
    'StadiumDirtClip': 362,
    'StadiumGrass': 363,
}



import math

class CGameHeader(object):
    """A generic header class that contains it's class ID."""
    def __init__(self, id):
        self.id = id


class CGameCtnCollectorList(object):
    """A header that holds a list of CollectorStock's."""
    def __init__(self, id):
        self.id = id
        self.stocks = []


class CollectorStock(object):
    """A header that holds a stock."""
    def __init__(self, block_name, collection, author):
        self.block_name = block_name
        self.collection = collection
        self.author = author


class MapBlock(object):
    """A header that holds information about a specific block contained within the Challenge data."""
    def __init__(self):
        self.name = None
        self.rotation = 0
        self.position = Vector3()
        self.speed = 0
        self.flags = 0
        self.params = 0
        self.skin_author = None
        self.skin = 0

    # 0b1000000000000
    # (flags & 0x1000) != 0     is on terrain
    # (flags & 0x1) != 0        connected once with another RoadMain
    # (flags & 0x2) != 0        connected twice with blocks (curve line)
    # (flags & 0x3) != 0        connected twice with blocks (straight line)
    # (flags & 0x4) != 0        connected three times with blocks
    # (flags & 0x5) != 0        connected four times with blocks
    # 6?
    # (flags & 0x7) == 0        not connected to any blocks
    def __str__(self):
        return (
            'Name: {}\n'
            'Rotation: {}\n'
            'Position: {}\n'
            'Flags: {}\n'
        ).format(self.name, self.rotation, self.position.as_array(), bin(self.flags))

class Vector3(object):
    """The Vector3 class represents a 3D vector, usually read directly from the GBX file."""
    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z

    def __add__(self, other):
        return Vector3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Vector3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        return None

    def __eq__(self, other):
        if isinstance(other, list):
            return self.x == other[0] and self.y == other[1] and self.z == other[2]

        return self.x == other.x and self.y == other.y and self.z == other.z

    def as_array(self):
        """Returns the vector as a list.
        
        Returns:
            the vector as a list made of 3 elements
        """
        return [self.x, self.y, self.z]


class CGameChallenge(CGameHeader):
    """A header that contains data related to the CGameChallenge class."""
    def __init__(self, id):
        self.id = id
        self.times = {}
        self.map_uid = None
        self.environment = None
        self.map_author = None
        self.map_name = None
        self.mood = None
        self.env_bg = None
        self.env_author = None
        self.map_size = ()
        self.flags = 0
        self.req_unlock = 0
        self.blocks = []
        self.items = []
        self.password_hash = None
        self.password_crc = None


class CGameBlockItem(CGameHeader):
    """A header that contains data related to the CGameBlockItem class."""
    def __init__(self):
        self.id = id
        self.path = None
        self.collection = None
        self.author = None
        self.waypoint = None
        self.position = Vector3()
        self.rotation = 0.0


class CGameWaypointSpecialProperty(CGameHeader):
    """A header that contains data related to the CGameWaypointSpecialProperty class."""
    def __init__(self, id):
        self.id = id
        self.tag = None
        self.spawn = 0
        self.order = 0


class CGameCommon(CGameHeader):
    """A header that contains data related to the CGameCommon class."""
    def __init__(self, id):
        self.id = id
        self.track_name = None


class CGameReplayRecord(CGameCommon):
    """A header that contains data related to the CGameReplayRecord class."""
    def __init__(self, id):
        self.id = id
        self.track = None
        self.nickname = None
        self.driver_login = None


class CGameGhost(CGameHeader):
    """A header that contains data related to the CGameGhost class."""
    def __init__(self, id):
        self.id = id
        self.records = []
        self.sample_period = None


class CGameCtnGhost(CGameGhost):
    """A header that contains data related to the CGameCtnGhost class."""
    def __init__(self, id):
        self.id = id
        self.race_time = 0
        self.num_respawns = 0
        self.light_trail_color = Vector3()
        self.stunts_score = 0
        self.uid = None
        self.login = None
        self.cp_times = []
        self.control_entries = []
        self.game_version = ''
        self.control_names = []
        self.events_duration = 0
        self.is_maniaplanet = False
        super(CGameCtnGhost, self).__init__(id)

class ControlEntry(object):
    """A header that contains data related to the control entries contained within the CGameCtnGhost class."""
    def __init__(self, time, event_name, enabled, flags):
        self.time = time
        self.event_name = event_name
        self.enabled = enabled
        self.flags = flags

class GhostSampleRecord(object):
    """A header that contains a single sample out of the ghost data such as position, rotation of the car and more."""
    BLOCK_SIZE_XZ = 32
    BLOCK_SIZE_Y = 8

    def __init__(self, position, angle, axis_heading, axis_pitch, speed, vel_heading, vel_pitch):
        self.position = position
        self.angle = angle
        self.axis_heading = axis_heading
        self.axis_pitch = axis_pitch
        self.speed = speed
        self.vel_heading = vel_heading
        self.vel_pitch = vel_pitch

    @property
    def display_speed(self):
        """Returns the display speed, which would be shown in the race.

        Returns:
            an integer from 0 to 1000 that represents the display speed"""
        if self.speed == 0x8000:
            return 0

        return int(abs(math.exp(self.speed / 1000.0) * 3.6))

    def get_block_position(self, xoff=0, yoff=0, zoff=0):
        """Calculates the block coordinates that the car is currently passing through in this sample record.

        Returns:
            a Vector3 containing the block coordinates
        """
        x = int((self.position.x + xoff) / self.BLOCK_SIZE_XZ)
        y = int((self.position.y + yoff) / self.BLOCK_SIZE_Y)
        z = int((self.position.z + zoff) / self.BLOCK_SIZE_XZ)
        return Vector3(x, y, z)



import logging
from enum import IntEnum

import lzo
import zlib

import pygbx.headers as headers
from pygbx.bytereader import ByteReader

class GbxType(IntEnum):
    """Represents the type of the main or auxiliary class contained within the GBX file.

    Some classes may have a different or multiple ID's, depending on what version of the GBX file is being parsed.
    """
    CHALLENGE = 0x03043000
    CHALLENGE_OLD = 0x24003000
    COLLECTOR_LIST = 0x0301B000
    CHALLENGE_PARAMS = 0x0305B000
    BLOCK_SKIN = 0x03059000
    WAYPOINT_SPECIAL_PROP = 0x0313B000
    ITEM_MODEL = 0x2E002000
    REPLAY_RECORD = 0x03093000
    REPLAY_RECORD_OLD = 0x02407E000
    GAME_GHOST = 0x0303F005
    CTN_GHOST = 0x03092000
    CTN_GHOST_OLD = 0x2401B000 
    CTN_COLLECTOR = 0x0301A000
    CTN_OBJECT_INFO = 0x0301C000
    CTN_DECORATION = 0x03038000
    CTN_COLLECTION = 0x03033000
    GAME_SKIN = 0x03031000
    GAME_PLAYER_PROFILE = 0x0308C000
    MW_NOD = 0x01001000
    UNKNOWN = 0x0

class GbxLoadError(Exception):
    """Thrown when the Gbx class fails to parse the provided Gbx object"""
    def __init__(self, message):
        """Initializes the Exception instance with a message.

        Args:
            message (str): the message
        """
        self.message = message


class Gbx(object):
    """The Gbx class provides the main interface for parsing GBX files and retrieving data
    that is contained within these files. The class provides support primarily for parsing Challenges and Replays.
    The class should properly support most TMNF/TMUF files, with support for some TM2 files.

    It is not guaranteed that the class will parse all of the data in a GBX file. If an error is encountered,
    an error will be logged with the logging module, but the class will attempt to read until the end of the file, unless
    the file is not a GBX file, then a GbxLoadError is thrown.

    The class uses the ByteReader class which provides support for reading data types exposed by the GBX file format,
    found on https://wiki.xaseco.org/wiki/GBX.

    If the class does not provide support for reading a chunk you want to specifically read, use find_raw_chunk_id
    to obtain access to a ByteReader with the cursor positioned at the beggining of provided chunk ID.
    """

    def __init__(self, obj):
        """Creates the main Gbx instance from a file path or from bytes object.

        Parses the Gbx file sequentially, reading all supported chunks until no more chunks have
        been found. Parsing can fail depending on the what classes or chunks it contains and what
        version of the GBX file is being parsed. 

        Args:
            obj (str/bytes): a file path to the Gbx file or bytes object containing the Gbx data

        Raises:
            GbxLoadError: raised when the supplied object is not a GBX file or data
        """
        if isinstance(obj, str):
            self.f = open(obj, 'rb')
            self.root_parser = ByteReader(self.f)
        else:
            self.root_parser = ByteReader(obj)

        self.magic = self.root_parser.read(3, '3s')
        if self.magic.decode('utf-8') != 'GBX':
            raise GbxLoadError(f'obj is not a valid Gbx data: magic string is incorrect')
        self.version = self.root_parser.read(2, 'H')
        self.classes = {}
        self.root_classes = {}
        self.positions = {}
        self.__current_class = None
        self.__current_waypoint = None
        self.__replay_header_info = {}

        self.root_parser.skip(3)
        if self.version >= 4:
            self.root_parser.skip(1)

        if self.version >= 3:
            self.class_id = self.root_parser.read_uint32()
            try:
                self.type = GbxType(self.class_id)
            except ValueError:
                self.type = GbxType.UNKNOWN

            if self.version >= 6:
                self._read_user_data()

            self.num_nodes = self.root_parser.read_uint32()

        self.num_external_nodes = self.root_parser.read_uint32()
        if self.num_external_nodes > 0:
            self.root_parser.read_uint32()
            self.__read_sub_folder()
            for node in range(self.num_external_nodes):
                flags = self.root_parser.read_uint32()
                if (flags & 4) == 0:
                    self.root_parser.read_string()
                else:
                    self.root_parser.read_uint32()

                self.root_parser.skip(4)
                if self.version >= 5:
                    self.root_parser.skip(4)

                if (flags & 4) == 0:
                    self.root_parser.skip(4)

        self.root_parser.push_info()
        self.positions['data_size'] = self.root_parser.pop_info()

        data_size = self.root_parser.read_uint32()
        compressed_data_size = self.root_parser.read_uint32()
        cdata = self.root_parser.read(compressed_data_size)
        self.data = bytearray(lzo.decompress(cdata, False, data_size))

        bp = ByteReader(self.data)
        self._read_node(self.class_id, -1, bp)

    def __read_sub_folder(self):
        num_sub_folders = self.root_parser.read_uint32()
        for folder in range(num_sub_folders):
            self.root_parser.read_string()
            self.__read_sub_folder()

    """Finds a raw chunk ID in the file, skipping through any data that does not match the chunk ID provided.

    It is not guaranteed that the chunk found is indeed the desired data, as it could be other unrelated
    chunk that bytes happened to form the chunk ID provided.

    Args:
        chunk_id (int): the chunk ID to search for

    Returns:
        ByteParser with the current position set right after the chunk ID, or None
        if no specified chunk ID was found
    """
    def find_raw_chunk_id(self, chunk_id):
        bp = ByteReader(self.data[:])
        for i in range(len(self.data) - 4):
            bp.pos = i
            if bp.read_uint32() == chunk_id:
                return bp

        return None

    def get_class_by_id(self, class_id):
        """Finds the header that corresponds to the provided class ID.
        
        This is a wrapper around get_classes_by_ids that returns the first
        element if any headers have been found for a single ID. See get_classes_by_ids for more details.

        Args:
            class_id (int): the class ID to be retrieved
        
        Returns:
            a single instance of the header corresponding to the provided class ID
        """
        classes = self.get_classes_by_ids([class_id])
        if len(classes) == 0:
            return None

        return classes[0]

    def get_classes_by_ids(self, class_ids):
        """Retrieves all headers that match any of the class ID's in the provided array.

        Searches through the list of parsed headers and returns an array of headers that match
        any of the provided ID's in the array. See GbxType for predefined class ID's that are supported.

        Args:
            class_ids (list): the class ID's that have to be matched

        Returns:
            a list of matched headers
        """        
        classes = []
        for cl in list(self.classes.values()) + list(self.root_classes.values()):
            if cl.id in class_ids:
                classes.append(cl)

        return classes

    def _read_user_data(self):
        entries = {}

        self.root_parser.push_info()
        self.user_data_size = self.root_parser.read_uint32()
        self.positions['user_data_size'] = self.root_parser.pop_info()

        user_data_pos = self.root_parser.pos
        num_chunks = self.root_parser.read_uint32()
        for _ in range(num_chunks):
            if self.root_parser.pos >= self.root_parser.size - 1:
                self.root_parser.pos = user_data_pos + self.user_data_size
                return

            cid = self.root_parser.read_uint32()
            self.root_parser.push_info()
            size = self.root_parser.read_uint32()
            self.positions[str(cid)] = self.root_parser.pop_info()
            entries[cid] = size

        for cid, size in entries.items():
            self._read_header_entry(cid, size)

        self.root_parser.pos = user_data_pos + self.user_data_size

    def _read_header_entry(self, cid, size):
        if cid == 0x03043002 or cid == 0x24003002:
            version = self.root_parser.read_byte()
            if version < 3:
                for _ in range(3):
                    self.root_parser.read_string_lookback()
                self.root_parser.read_string()

            self.root_parser.skip(4)
            if version >= 1:
                self.root_parser.skip(16)
                if version == 2:
                    self.root_parser.skip(4)

                if version >= 4:
                    self.root_parser.skip(4)
                    if version >= 5:
                        self.root_parser.skip(4)

                        if version == 6:
                            self.root_parser.skip(4)
                        
                        if version >= 7:
                            self.root_parser.read_uint32()

                            if version >= 9:
                                self.root_parser.skip(4)

                                if version >= 10:
                                    self.root_parser.skip(4)

                                    if version >= 11:
                                        self.root_parser.skip(4)
                                        
                                        if version >= 12:
                                            self.root_parser.skip(4)

                                            if version >= 13:
                                                self.root_parser.skip(8)

        elif cid == 0x03043003 or cid == 0x24003003:
            p = self.root_parser.pos
            self.root_parser.read_byte()
            for _ in range(3):
                self.root_parser.read_string_lookback()

            game_class = headers.CGameCommon(cid)

            self.root_parser.push_info()
            game_class.track_name = self.root_parser.read_string()
            self.positions['track_name'] = self.root_parser.pop_info()

            self.root_parser.read_byte()

            self.root_classes[cid] = game_class

            self.root_parser.pos = p + size
        elif cid == 0x03043005 or cid == 0x24003005:
            self.__community = self.root_parser.read_string()
        elif cid == 0x03093000 or cid == 0x2403F000:
            version = self.root_parser.read_uint32()
            self.__replay_header_info['version'] = version
            if version >= 2:
                for _ in range(3):
                    self.root_parser.read_string_lookback()
                self.root_parser.skip(4)
                self.__replay_header_info['nickname'] = self.root_parser.read_string()
                if version >= 6:
                    self.__replay_header_info['driver_login'] = self.root_parser.read_string()
                    self.root_parser.skip(1)
                    self.root_parser.read_string_lookback()
        elif cid == 0x03093002 or cid == 0x2403F002:
            self.root_parser.skip(8)
            for _ in range(4):
                self.root_parser.read_string()
        else:
            self.root_parser.skip(size)

    def _read_node(self, class_id, depth, bp, add=True):
        oldcid = 0
        cid = 0

        if class_id == GbxType.CHALLENGE or class_id == GbxType.CHALLENGE_OLD:
            game_class = headers.CGameChallenge(class_id)
            if hasattr(self, '__community'):
                game_class.community = self.__community
        elif class_id == GbxType.REPLAY_RECORD or class_id == GbxType.REPLAY_RECORD_OLD:
            game_class = headers.CGameReplayRecord(class_id)
            if 'nickname' in self.__replay_header_info:
                game_class.nickname = self.__replay_header_info['nickname']
            if 'driver_login' in self.__replay_header_info:
                game_class.driver_login = self.__replay_header_info['driver_login']

        elif class_id == GbxType.WAYPOINT_SPECIAL_PROP or class_id == 0x2E009000:
            game_class = headers.CGameWaypointSpecialProperty(class_id)
            self.__current_waypoint = game_class
            add = False
        elif class_id == GbxType.CTN_GHOST or class_id == GbxType.CTN_GHOST_OLD:
            game_class = headers.CGameCtnGhost(class_id)
        elif class_id == GbxType.GAME_GHOST:
            game_class = headers.CGameGhost(class_id)
        elif class_id == GbxType.COLLECTOR_LIST:
            game_class = headers.CGameCtnCollectorList(class_id)
        else:
            game_class = headers.CGameHeader(class_id)

        self.__current_class = game_class

        if add:
            self.classes[depth] = game_class

        while True:
            oldcid = cid
            cid = bp.read_uint32()
            logging.debug(f'Reading chunk {hex(cid)}')

            if cid == 0xFACADE01:
                logging.debug('Encountered chunk 0xFACADE01, stopping')
                break

            skipsize = -1
            skip = bp.read_int32()
            if skip == 0x534B4950:
                skipsize = bp.read_uint32()
            else:
                bp.pos -= 4

            if cid == 0x0304300D or cid == 0x2400300D:
                bp.read_string_lookback()
                bp.read_string_lookback()
                bp.read_string_lookback()
            elif cid == 0x03043011 or cid == 0x24003011:
                for _ in range(2):
                    idx = bp.read_int32()
                    if idx >= 0 and idx not in self.classes:
                        _class_id = bp.read_int32()
                        self._read_node(_class_id, idx, bp)
                bp.read_uint32()
            elif cid == 0x301B000 or cid == 0x2403C000:
                itemsct = bp.read_uint32()
                for _ in range(itemsct):
                    bp.read_string_lookback()
                    bp.read_string_lookback()
                    bp.read_string_lookback()
                    bp.read_uint32()
            elif cid == 0x0305B000 or cid == 0x2400C000:
                bp.skip(8 * 4)
            elif cid == 0x0305B001 or cid == 0x2400C001:
                bp.read_string()
                bp.read_string()
                bp.read_string()
                bp.read_string()
            elif cid == 0x0305B004 or cid == 0x2400C004:
                game_class.times = {
                    'bronze': bp.read_int32(),
                    'silver': bp.read_int32(),
                    'gold': bp.read_int32(),
                    'author': bp.read_int32()
                }

                bp.read_uint32()
            elif cid == 0x0305B005 or cid == 0x2400C005:
                bp.skip(4 * 3)
            elif cid == 0x0305B006 or cid == 0x2400C006:
                count = bp.read_uint32()
                bp.skip(count * 4)
            elif cid == 0x0305B008 or cid == 0x2400C008:
                bp.skip(2 * 4)
            elif cid == 0x0305B00A:
                bp.skip(9 * 4)                
            elif cid == 0x0305B00D:
                bp.skip(1 * 4)
            elif cid == 0x03043014 or cid == 0x03043029:
                bp.read(16 + 4)
                # m = hashlib.md5()
                # m.update(bp.read(16))
                # if isinstance(self.__current_class, headers.CGameChallenge):
                    # self.__current_class.password_hash = m.hexdigest()
                    # self.__current_class.password_crc = bp.read_uint32()
            elif cid == 0x03043017:
                num_cps = bp.read_uint32()
                for _ in range(num_cps):
                    bp.read_uint32()
                    bp.read_uint32()
                    bp.read_uint32()

            elif cid == 0x0304301F or cid == 0x2400301F:
                game_class.map_uid = bp.read_string_lookback()
                game_class.environment = bp.read_string_lookback()
                game_class.map_author = bp.read_string_lookback()

                bp.push_info()
                game_class.map_name = bp.read_string()
                self.positions['map_name'] = bp.pop_info()

                bp.push_info()
                game_class.mood = bp.read_string_lookback()
                self.positions['mood'] = bp.pop_info()
                game_class.env_bg = bp.read_string_lookback()
                game_class.env_author = bp.read_string_lookback()

                game_class.map_size = (
                    bp.read_int32(),
                    bp.read_int32(),
                    bp.read_int32()
                )

                game_class.req_unlock = bp.read_int32()
                game_class.flags = bp.read_int32()

                bp.push_info()
                num_blocks = bp.read_uint32()
                i = 0
                while i < num_blocks:
                    block = headers.MapBlock()
                    block.name = bp.read_string_lookback()
                    if block.name != 'Unassigned1':
                        game_class.blocks.append(block)
                    block.rotation = bp.read_byte()
                    block.position = headers.Vector3(
                        bp.read_byte(),
                        bp.read_byte(),
                        bp.read_byte()
                    )

                    if game_class.flags > 0:
                        block.flags = bp.read_uint32()
                    else:
                        block.flags = bp.read_uint16()

                    if block.flags == 0xFFFFFFFF:
                        continue

                    if (block.flags & 0x8000) != 0:
                        block.skin_author = bp.read_string_lookback()
                        if game_class.flags >= 6:
                            # TM2 flags
                            bp.read_string() # Block waypoint type {Spawn, Goal}
                            bp.read_int32()
                            self._read_node(0x2E009000, 0, bp, False)
                        else:
                            block.skin = bp.read_int32()
                            if block.skin >= 0 and block.skin not in self.classes:
                                _class_id = bp.read_int32()
                                self._read_node(_class_id, block.skin, bp)

                        if (block.flags & 0x100000) != 0:
                            block.params = bp.read_int32()
                            if block.params >= 0 and block.params not in self.classes:
                                _class_id = bp.read_int32()
                                self._read_node(_class_id, block.params, bp)

                    i += 1

                self.positions['block_data'] = bp.pop_info()              
            elif cid == 0x03043022:
                bp.skip(4)
            elif cid == 0x03043024:
                version = bp.read_byte()
                if version >= 3:
                    bp.skip(32)

                file_path = bp.read_string()
                if len(file_path) > 0 or version >= 3:
                    bp.read_string()
            elif cid == 0x03043025:
                bp.skip(4 * 4)
            elif cid == 0x03043026:
                idx = bp.read_int32()
                if idx >= 0 and idx not in self.classes:
                    _class_id = bp.read_int32()
                    self._read_node(_class_id, idx, bp)
            # elif cid == 0x03043028:
            #     archive_gm_cam_val = bp.read_int32()
            #     if archive_gm_cam_val == 1:
            #         bp.skip(1 + 4 * 7)

            #     bp.read_string()
            elif cid == 0x0304302A:
                bp.read_int32()
            elif cid == 0x3043040:
                bp.pos -= 4
                bp.push_info()
                bp.pos += 4

                item_bp = ByteReader(bp.data)
                item_bp.pos = bp.pos

                item_bp.skip(2 * 4)

                item_bp.push_info()
                item_bp.skip(2 * 4)

                num_items = item_bp.read_uint32()
                for i in range(num_items):
                    item_bp.skip(4 * 3)
                    item = headers.CGameBlockItem()
                    item.path = item_bp.read_string_lookback()
                    item.collection = item_bp.read_string_lookback()
                    item.author = item_bp.read_string_lookback()
                    item.rotation = item_bp.read_float()
                    item_bp.skip(15)

                    item.position = item_bp.read_vec3()

                    idx = item_bp.read_int32()
                    if idx >= 0:
                        self._read_node(0x2E009000, idx, item_bp)

                    item.waypoint = self.__current_waypoint
                    self.__current_waypoint = None

                    item_bp.skip(4 * 4 + 2)

                    self._read_node(0x3101004, 0, item_bp, add=False)

                    game_class.items.append(item)

                item_bp.skip(4)
                bp.pos = item_bp.pos
            elif cid == 0x03059002 or cid == 0x2403A002:
                bp.read_string()
                for i in range(2):
                    version = bp.read_byte()
                    if version >= 3:
                        bp.skip(32)

                    file_path = bp.read_string()
                    # (?) according to https://wiki.xaseco.org/wiki/GBX
                    # we need to check if the file path is not empty
                    # here but it crashes reading the file for TM2 challenges
                    if len(file_path) > 0 and version >= 1:
                        bp.read_string()
            elif cid == 0x03043022:
                bp.read_int32()
            elif cid == 0x03043024:
                bp.read_byte()
                bp.skip(32)
                bp.read_string()
                bp.read_string()
            elif cid == 0x03043025:
                bp.skip(16)
            elif cid == 0x03043026:
                idx = bp.read_int32()
                if idx >= 0 and idx not in self.classes:
                    _class_id = bp.read_int32()
                    self._read_node(_class_id, idx, bp)
            elif cid == 0x03043028:
                p = bp.read_int32()
                if p != 0:
                    bp.skip(1 + 4 * 3 * 3 + 4 * 3 + 4 + 4 + 4)

                bp.read_string()
            elif cid == 0x0304302A:
                bp.read_int32()
            elif cid == GbxType.WAYPOINT_SPECIAL_PROP or cid == 0x2E009000:
                version = bp.read_uint32()
                if version == 1:
                    game_class.spawn = bp.read_uint32()
                    game_class.order = bp.read_uint32()
                elif version == 2:
                    game_class.tag = bp.read_string()
                    game_class.order = bp.read_uint32()
            elif cid == 0x03059000:
                bp.read_string()
                bp.read_string()
            elif cid == 0x0303F005:
                Gbx.read_ghost(game_class, bp)
            elif cid == 0x0303F006:
                bp.skip(4)
                Gbx.read_ghost(game_class, bp)
            elif cid == 0x03093002 or cid == 0x2403F002:
                map_gbx_size = bp.read_uint32()
                data = bytes(bp.read(map_gbx_size))
                try:
                    game_class.track = Gbx(data)
                except Exception as e:
                    logging.error(f'Failed to parse map data: {e}')

            elif cid == 0x03093007:
                bp.skip(4)
            elif cid == 0x03093014 or cid == 0x2403F014:
                bp.skip(4)
                num_ghosts = bp.read_uint32()
                for _ in range(num_ghosts):
                    idx = bp.read_int32()
                    if idx >= 0 and idx not in self.classes:
                        _class_id = bp.read_uint32()
                        self._read_node(_class_id, idx, bp)
                
                bp.skip(4)
                # num_extras = bp.read_uint32()
                # bp.skip(num_extras * 8)
            elif cid == 0x3093015:
                idx = bp.read_int32()
                if idx >= 0 and idx not in self.classes:
                    _class_id = bp.read_uint32()
                    self._read_node(_class_id, idx, bp)

            elif cid == 0x03043021 or cid == 0x24003021:
                for _ in range(3):
                    idx = bp.read_int32()
                    if idx >= 0 and idx not in self.classes:
                        _class_id = bp.read_int32()
                        self._read_node(_class_id, idx, bp)
            elif cid == 0x03043022 or cid == 0x24003022:
                bp.skip(4)
            elif cid == 0x03043024 or cid == 0x24003024:
                version = bp.read_byte()
                if version >= 3:
                    bp.skip(32)
                
                path = bp.read_string()
                if len(path) > 0 and version >= 1:
                    bp.read_string()
            elif cid == 0x03043025 or cid == 0x24003025:
                bp.skip(4 * 4)
            elif cid == 0x03043026 or cid == 0x24003026:
                idx = bp.read_int32()
                if idx >= 0 and idx not in self.classes:
                    _class_id = bp.read_int32()
                    self._read_node(_class_id, idx, bp)
            elif cid == 0x03092005 or cid == 0x2401B005:
                game_class.race_time = bp.read_uint32()
            elif cid == 0x03092008 or cid == 0x2401B008:
                game_class.num_respawns = bp.read_uint32()
            elif cid == 0x03092009 or cid == 0x2401B009:
                game_class.light_trail_color = bp.read_vec3()
            elif cid == 0x0309200A or cid == 0x2401B00A:
                game_class.stunts_score = bp.read_uint32()
            # The GBX spec is wrong here.
            # 0x0309200B contains how many CP times there are
            # after that, there is a list of (uint32, uint32)
            # tuples, the first element is the time, the second
            # is unknown
            elif cid == 0x0309200B or cid == 0x2401B00B:
                num = bp.read_uint32()
                cp_times = []
                for i in range(num):
                    cp_times.append(bp.read_uint32())
                    bp.skip(4)
                
                game_class.cp_times = cp_times
            elif cid == 0x309200C or cid == 0x2401B00C:
                bp.skip(4)
            elif cid == 0x309200E or cid == 0x2401B00E:
                game_class.uid = bp.read_string_lookback()

                # For TM2
                if ('version' in self.__replay_header_info
                    and self.__replay_header_info['version'] >= 8):
                    pos = bp.pos
                    try:
                        game_class.login = bp.read_string()
                    except:
                        bp.pos = pos

            elif cid == 0x309200F or cid == 0x2401B00F:
                game_class.login = bp.read_string()
            elif cid == 0x3092010 or cid == 0x2401B010:
                bp.read_string_lookback()
            elif cid == 0x3092012 or cid == 0x2401B012:
                bp.skip(4 + 16)
                # import binascii
                # # bp.read(4)
                # print(bp.read_uint32())
                # print(f'{binascii.hexlify(bp.read(16))}')
                # print()
            elif cid == 0x3092013 or cid == 0x2401B013:
                bp.skip(4 + 4)
            elif cid == 0x3092014 or cid == 0x2401B014:
                bp.skip(4)
            elif cid == 0x3092015 or cid == 0x2401B015:
                bp.read_string_lookback()
            elif cid == 0x3092018 or cid == 0x2401B018:
                bp.read_string_lookback()
                bp.read_string_lookback()
                bp.read_string_lookback()
            elif cid == 0x3092019 or cid == 0x03092025 or cid == 0x2401B019 or cid == 0x2401B011:
                Gbx.read_ghost_events(game_class, bp, cid)
            elif cid == 0x309201c:
                bp.skip(32)
            elif cid == 0x03093004 or cid == 0x2403f004:
                bp.skip(4 * 4)
            elif skipsize != -1:
                bp.skip(skipsize)
                cid = oldcid
            else:
                return

    @staticmethod
    def read_ghost_events(game_class, bp, cid):
        if cid == 0x03092025:
            game_class.is_maniaplanet = True
            bp.skip(4)

        game_class.events_duration = bp.read_uint32()
        if game_class.events_duration != 0:
            bp.skip(4)

            num_control_names = bp.read_uint32()
            game_class.control_names = []
            for _ in range(num_control_names):
                name = bp.read_string_lookback()
                if name != '':
                    game_class.control_names.append(name)
                
            if len(game_class.control_names) == 0:
                return

            num_control_entries = bp.read_uint32()
            bp.skip(4)
            for _ in range(num_control_entries):
                time = bp.read_uint32() - 100000
                name = game_class.control_names[bp.read_byte()]
                entry = headers.ControlEntry(time, name, bp.read_uint16(), bp.read_uint16()) 
                game_class.control_entries.append(entry)
            
            game_class.game_version = bp.read_string()
            bp.skip(3 * 4)
            bp.read_string()
            bp.skip(4)

    @staticmethod
    def read_ghost(game_class, bp):
        uncomp_sz = bp.read_uint32()
        comp_sz = bp.read_uint32()
        comp_data = bp.read(comp_sz)
        data = zlib.decompress(comp_data, 0, uncomp_sz)

        gr = ByteReader(data)
        gr.skip(3 * 4)
        game_class.sample_period = gr.read_uint32()
        gr.skip(1 * 4)

        sample_data_sz = gr.read_uint32()
        sample_data_pos = gr.pos
        gr.skip(sample_data_sz)

        sample_sizes = []
        num_samples = gr.read_uint32()
        fso = 0
        if num_samples > 0:
            fso = gr.read_uint32()
            if num_samples > 1:
                sps = gr.read_int32()
                if sps == -1:
                    sample_sizes = []
                    for _ in range(num_samples - 1):
                        sample_sizes.append(gr.read_uint32())
                else:
                    sample_sizes.append(sps)

        gr.pos = sample_data_pos
        gr.skip(fso)
        for i in range(num_samples):
            sample_pos = gr.pos

            record = headers.GhostSampleRecord(
                gr.read_vec3(), gr.read_uint16(), gr.read_int16(),
                gr.read_int16(), gr.read_int16(),
                gr.read_int8(), gr.read_int8())

            len_sizes = len(sample_sizes)
            if i >= len_sizes:
                if len_sizes >= 1:
                    sample_sz = sample_sizes[0]
                else:
                    sample_sz = 0
            else:
                sample_sz = sample_sizes[i]

            record.raw_data = gr.read(sample_sz - (gr.pos - sample_pos))
            # import binascii
            # print(f'{i} {binascii.hexlify(record.raw_data)}')
            game_class.records.append(record)



import logging
import struct
from io import IOBase
from pygbx.headers import Vector3
from os import SEEK_END

class PositionInfo(object):
    """
    This classes holds information that is mainly private to
    the Gbx class but can still be retrieved through the positions member.

    The PositionInfo marks a specific section in the file through it's position and size.
    """

    def __init__(self, pos, size):
        """Constructs a new PositionInfo"""

        self.pos = pos
        self.size = size

    @property
    def valid(self):
        """Checks if the instance of the section is valid
        
        Returns:
            True if the instance points to a valid section in the file, False otherwise
        """
        return self.pos > -1 and self.size > 0


class ByteReader(object):
    """The ByteReader class is used by the Gbx class to read specific data types supported by the GBX file format.

    The class provides convinience methods for reading raw types such as integers, strings and vectors, which
    are the main data types of the GBX file format. While reading the file, the Gbx class may instantiate multiple
    instances of ByteReader to read different parts of the file. This is because some chunks depend on the state
    of the reader, this state can be e.g: lookback strings.

    ByteReader accepts reading from raw bytes as well as from a file handle.
    """
    def __init__(self, obj):
        """Constructs a new ByteReader with the provided object.

        Args:
            obj (file/bytes): a file handle opened through open() or a bytes object
        """
        self.data = obj
        if isinstance(obj, IOBase):
            self.get_bytes = self.__get_bytes_file
            self.data.seek(0, SEEK_END)
            self.size = self.data.tell()
            self.data.seek(0)
        else:
            self.get_bytes = self.__get_bytes_generic
            self.size = len(self.data)

        self.pos = 0
        self.seen_loopback = False
        self.stored_strings = []
        self.current_info = PositionInfo(-1, 0)

    def push_info(self):
        """Begins a section that can be then retrieved with pop_info."""
        self.current_info = PositionInfo(self.pos, 0)


    def pop_info(self):
        """Ends the section began with push_info.
    
        Returns:
            a PositionInfo marking the section
        """
        self.current_info.size = self.pos - self.current_info.pos
        info = self.current_info
        self.current_info = PositionInfo(-1, 0)
        return info

    def read(self, num_bytes, typestr=None):
        """Reads an arbitrary amount of bytes from the buffer.

        Reads the buffer of length num_bytes and optionally
        takes a type string that is passed to struct.unpack if not None.

        Args:
            num_bytes (int): the number of bytes to read from the buffer
            typestr (str): the format character used by the struct module, passing None does not unpack the bytes
        
        Returns:
            the bytes object, if no type string was provided, type returned by struct.unpack otherwise
        """
        val = self.get_bytes(num_bytes)
        self.pos += num_bytes
        if typestr == None:
            return val
        try:
            return struct.unpack(typestr, val)[0]
        except Exception as e:
            logging.error(e)
            return 0

    def size(self):
        if isinstance(self.data, IOBase):
            return self.data.tell()
        else:
            return len(self.data)

    def __get_bytes_file(self, num_bytes):
        self.data.seek(self.pos)
        return self.data.read(num_bytes)

    def __get_bytes_generic(self, num_bytes):
        return self.data[self.pos:self.pos + num_bytes]

    def read_int32(self):
        """Reads a signed int32.
        
        Returns:
            the integer read from the buffer
        """
        return self.read(4, 'i')

    def read_uint32(self):
        """Reads an unsigned int32.
        
        Returns:
            the integer read from the buffer
        """
        return self.read(4, 'I')

    def read_int16(self):
        """Reads a signed int16.
        
        Returns:
            the integer read from the buffer
        """
        return self.read(2, 'h')

    def read_uint16(self):
        """Reads an unsigned int16.
        
        Returns:
            the integer read from the buffer
        """
        return self.read(2, 'H')

    def read_int8(self):
        """Reads a signed int8.
        
        Returns:
            the integer read from the buffer
        """
        return self.read(1, 'b')

    def read_float(self):
        """Reads a 32 bit float.
        
        Returns:
            the float read from the buffer
        """
        return self.read(4, 'f')

    def read_vec3(self):
        """Reads 12 bytes as 3 floats from the buffer and packs them into a Vector3.
        
        Returns:
            the vector read from the buffer
        """
        return Vector3(self.read_float(), self.read_float(), self.read_float())

    def read_string(self):
        """Reads a string from the buffer, first reading the length, then it's data.

        Returns:
            the string read from the buffer, None if there was an error
        """
        strlen = self.read_uint32()
        try:
            return self.read(strlen, str(strlen) + 's').decode('utf-8')
        except UnicodeDecodeError as e:
            logging.error(f'Failed to read string: {e}')
            return None

    def read_byte(self):
        """Reads a single byte from the buffer.

        Returns:
            the single byte read from the buffer
        """
        val = self.get_bytes(1)[0]
        self.pos += 1
        return val

    def skip(self, num_bytes):
        """Skips provided amount of bytes in the buffer

        Args:
            num_bytes (int): the number of bytes to skip
        """
        self.pos += num_bytes

    def read_string_lookback(self):
        """Reads a special string type in the GBX file format called the lookbackstring.
        
        Such type is used to reference already read strings, or introduce them if they were not 
        read yet. A ByteReader instance keeps track of lookback strings previously read and
        returns an already existing string, if the data references it. For more information,
        see the lookbackstring type in the GBX file format: https://wiki.xaseco.org/wiki/GBX.

        Returns:
            the lookback string read from the buffer
        """
        if not self.seen_loopback:
            self.read_uint32()

        self.seen_loopback = True
        inp = self.read_uint32()
        if (inp & 0xc0000000) != 0 and (inp & 0x3fffffff) == 0:
            s = self.read_string()
            self.stored_strings.append(s)
            return s

        if inp == 0:
            s = self.read_string()
            self.stored_strings.append(s)
            return s

        if inp == -1:
            return ''

        if (inp & 0x3fffffff) == inp:
            if inp == 11:
                return 'Valley'
            elif inp == 12:
                return 'Canyon'
            elif inp == 13:
                return 'Lagoon'
            elif inp == 17:
                return 'TMCommon'
            elif inp == 202:
                return 'Storm'
            elif inp == 299:
                return 'SMCommon'
            elif inp == 10003:
                return 'Common'

        inp &= 0x3fffffff
        if inp - 1 >= len(self.stored_strings):
            return ''
        return self.stored_strings[inp - 1]



LAGOON_BLOCKS = {
    'RoadMainSmall': 1,
    'RoadMain': 2,
    'RoadMainStart': 3,
    'RoadMainCheckpoint': 4,
    'RoadMainFinish': 5,
    'RoadMainMultilap': 6,
    'RoadMainTurbo': 7,
    'RoadMainTurbo2': 8,
    'RoadMainFW': 9,
    'RoadMainToll2': 10,
    'RoadMainToll3': 11,
    'RoadMainObstacleBump': 12,
    'RoadMainRamp': 13,
    'RoadMainRamp2': 14,
    'RoadMainBirds': 15,
    'RoadMainFreeSidewalk': 16,
    'RoadMainGTCurve2': 17,
    'RoadMainGTCurve3': 18,
    'RoadMainGTDiag2': 19,
    'RoadMainGTDiag2Mirror': 20,
    'RoadMainGTDiag3': 21,
    'RoadMainGTDiag3Mirror': 22,
    'RoadMainYShapedDiag': 23,
    'RoadMainYShapedCurve': 24,
    'RoadMainSlopeBase': 25,
    'RoadMainSlopeBase2': 26,
    'RoadMainBiSlopeStart': 27,
    'RoadMainBiSlopeEnd': 28,
    'RoadMainSlopeUTop': 29,
    'RoadMainSlopeUBottom': 30,
    'RoadMainSlopeBaseGT2Mirror': 31,
    'RoadMainSlopeBaseGT2': 32,
    'RoadMainSlopeStraight': 33,
    'RoadMainCheckpointUp': 34,
    'RoadMainCheckpointDown': 35,
    'RoadMainTurboUp': 36,
    'RoadMainTurboDown': 37,
    'RoadMainTurbo2Up': 38,
    'RoadMainTurbo2Down': 39,
    'RoadMainFWSlope': 40,
    'RoadMainBridgeOnSeaCliff13': 41,
    'RoadMainBridgeOnLandCliff12': 42,
    'RoadMainBridgeOnHill6': 43,
    'RoadMainBridgeOnHill': 44,
    'RoadMainSlopeBase2Hill': 45,
    'RoadMainBridgeOnBeach': 46,
    'ArenaToRoadMain': 47,
    'RoadThemeParkToRoadMain': 48,
    'HighWayToRoadMainSlope2Start': 49,
    'HighWayToRoadMainSlope2Start2x1': 50,
    'HighWayToRoadMainSlope2End': 51,
    'HighWayToRoadMainSlope2End2x1': 52,
    'HighWaySlope2UTop': 53,
    'HighWaySlope2UBottom': 54,
    'HighWaySlope2LoopLeft': 55,
    'HighWaySlope2LoopRight': 56,
    'HighWaySlope2Straight': 57,
    'HighWaySlope2CheckpointUp': 58,
    'HighWaySlope2CheckpointDown': 59,
    'HighWaySlope2TurboUp': 60,
    'HighWaySlope2TurboDown': 61,
    'HighWaySlope2Turbo2Up': 62,
    'HighWaySlope2Turbo2Down': 63,
    'HighWaySlope2-FW': 64,
    'HighWaySlope2CornerDownLeft': 65,
    'HighWaySlope2CornerDownRight': 66,
    'HighWaySlope2CornerUpLeft': 67,
    'HighWaySlope2CornerUpRight': 68,
    'HighWaySlope2GTCurve2DownLeft': 69,
    'HighWaySlope2GTCurve2DownRight': 70,
    'HighWaySlope2GTCurve2UpLeft': 71,
    'HighWaySlope2GTCurve2UpRight': 72,
    'HighWaySlope2ExchangeGTCurve2UpLeft': 73,
    'HighWaySlope2ExchangeGTCurve2UpRight': 74,
    'HighWaySlope2ExchangeGTCurve2DownLeft': 75,
    'HighWaySlope2ExchangeGTCurve2DownRight': 76,
    'HighWayTilt2Straight': 77,
    'HighWaySlope2CheckpointLeft': 78,
    'HighWaySlope2CheckpointRight': 79,
    'HighWaySlope2TurboLeft': 80,
    'HighWaySlope2TurboRight': 81,
    'HighWaySlope2Turbo2Left': 82,
    'HighWaySlope2Turbo2Right': 83,
    'HighWaySlope2-FWTilt': 84,
    'FreeWay': 85,
    'FreeWayStart': 86,
    'FreeWayCheckpoint': 87,
    'FreeWayFinish': 88,
    'FreeWayMultilap': 89,
    'FreeWayTurbo': 90,
    'FreeWayTurbo2': 91,
    'FreeWayFW': 92,
    'FreeWayBirds': 93,
    'FreeWayRamp': 94,
    'FreeWayRamp2': 95,
    'FreeWayCornerSign': 96,
    'FreeWayGTCurve2': 97,
    'FreeWayGTCurve3': 98,
    'FreeWayGTDiag2': 99,
    'FreeWayGTDiag2Mirror': 100,
    'FreeWayGTDiag3': 101,
    'FreeWayGTDiag3Mirror': 102,
    'FreeWayYShapedDiag': 103,
    'FreeWayYShapedCurve': 104,
    'FreeWayTransitionToRoadMainStraight': 105,
    'FreeWayTransitionToRoadMainSlopeBase': 106,
    'FreeWayTransitionToRoadMainSlopeBase2': 107,
    'FreeWayTransitionToRoadMainBiSlopeStart': 108,
    'FreeWayHillStraight': 109,
    'RoadThemePark': 110,
    'RoadThemeParkGTCurve2': 111,
    'RoadThemeParkGTDiag2': 112,
    'RoadThemeParkGTDiag2Mirror': 113,
    'RoadThemeParkStart': 114,
    'RoadThemeParkCheckpoint': 115,
    'RoadThemeParkFinish': 116,
    'RoadThemeParkMultilap': 117,
    'RoadThemeParkTurbo': 118,
    'RoadThemeParkTurbo2': 119,
    'RoadThemeParkFreeWheeling': 120,
    'ArenaToRoadThemeParkUp': 121,
    'RoadThemeParkTwistRoadtoWall2': 122,
    'RoadThemeParkTwistRoadtoWall2Mirror': 123,
    'RoadThemeParkTwistRoadtoWallGT2': 124,
    'RoadThemeParkTwistRoadtoWallGT2Mirror': 125,
    'RoadThemeParkLoopInStart': 126,
    'RoadThemeParkLoopInEnd': 127,
    'RoadThemeParkLoop': 128,
    'RoadThemeParkLoopMirror': 129,
    'RoadThemeParkWall-Straight': 130,
    'RoadThemeParkWall-Checkpoint': 131,
    'RoadThemeParkWall-FreeWheeling': 132,
    'RoadThemeParkWall-EndLeft': 133,
    'RoadThemeParkWall-EndRight': 134,
    'RoadThemeParkWallGTDiag2': 135,
    'RoadThemeParkWallGTDiag2Mirror': 136,
    'RoadThemeParkWallLoopIn': 137,
    'RoadThemeParkWallLoopOut': 138,
    'RoadThemeParkWallLoop': 139,
    'RoadThemeParkWallLoopMirror': 140,
    'RoadThemeParkWall-StraightUp': 141,
    'RoadThemeParkWall-CheckpointUp': 142,
    'RoadThemeParkWall-FreeWheelingUp': 143,
    'RoadThemeParkWall-StraightUp2': 144,
    'RoadThemeParkWall-StraightUp1': 145,
    'RoadThemeParkWall-EndUp': 146,
    'RoadThemeParkWall-Cross': 147,
    'RoadThemeParkWallUpGTDiag2': 148,
    'RoadThemeParkWallUpGTDiag2Mirror': 149,
    'RoadThemeParkWall-CornerDown': 150,
    'RoadThemeParkWall-CornerDownMirror': 151,
    'RoadThemeParkWall-CornerUp': 152,
    'RoadThemeParkWall-CornerUpMirror': 153,
    'RoadThemeParkWall-GTCurve2Down': 154,
    'RoadThemeParkWall-GTCurve2DownMirror': 155,
    'RoadThemeParkWall-GTCurve2Up': 156,
    'RoadThemeParkWall-GTCurve2UpMirror': 157,
    'RoadThemeParkTwistWallUp2': 158,
    'RoadThemeParkTwistWallUp2Mirror': 159,
    'RoadThemeParkTwistWallDownGT2': 160,
    'RoadThemeParkTwistWallDownGT2Mirror': 161,
    'RoadThemeParkTwistWallUpGT2': 162,
    'RoadThemeParkTwistWallUpGT2Mirror': 163,
    'RoadThemeParkLoopOutStart': 164,
    'RoadThemeParkLoopOutEnd': 165,
    'RoadThemeParkTwistWalltoFlip2': 166,
    'RoadThemeParkTwistWalltoFlip2Mirror': 167,
    'RoadThemeParkTwistWalltoFlipGT2': 168,
    'RoadThemeParkTwistWalltoFlipGT2Mirror': 169,
    'RoadThemeParkSuperLoopOutEnd': 170,
    'RoadThemeParkSuperLoopOutStart': 171,
    'RoadThemeParkWallSuperLoopOut': 172,
    'RoadThemeParkSuperTwistRoadtoWall2': 173,
    'RoadThemeParkSuperTwistRoadtoWall2Mirror': 174,
    'RoadThemeParkSuperTwistWallUp2': 175,
    'RoadThemeParkSuperTwistWallUp2Mirror': 176,
    'RoadThemeParkSuperTwistWalltoFlip2': 177,
    'RoadThemeParkSuperTwistWalltoFlip2Mirror': 178,
    'RoadThemeParkFlip': 179,
    'RoadThemeParkFlipFinish': 180,
    'RoadThemeParkFlip-Checkpoint': 181,
    'RoadThemeParkFlip-Turbo': 182,
    'RoadThemeParkFlip-Turbo2': 183,
    'RoadThemeParkFlip-FreeWheeling': 184,
    'RoadThemeParkFlipGTCurve2': 185,
    'RoadThemeParkFlipGTDiag2': 186,
    'RoadThemeParkFlipGTDiag2Mirror': 187,
    'ArenaBase': 188,
    'ArenaStart': 189,
    'ArenaCheckpoint': 190,
    'ArenaCheckpointX2': 191,
    'ArenaMultilap': 192,
    'ArenaMultilapX2': 193,
    'ArenaFinish': 194,
    'ArenaFinishX2': 195,
    'ArenaTurbo': 196,
    'ArenaTurbo2': 197,
    'ArenaFW': 198,
    'ArenaSlopeBase': 199,
    'ArenaSlopeBaseSupport': 200,
    'ArenaSlope2Start': 201,
    'ArenaSlope2StartSupport': 202,
    'ArenaSlope2End': 203,
    'ArenaSlope2EndSupport': 204,
    'ArenaSlope2CornerIn': 205,
    'ArenaSlope2GTCurve2': 206,
    'ArenaSlope2GTCurve3': 207,
    'ArenaSlope2Twist2Left': 208,
    'ArenaSlope2Twist2Right': 209,
    'ArenaSlope2Twist3Left': 210,
    'ArenaSlope2Twist3Right': 211,
    'ArenaSlope2Straight': 212,
    'ArenaSlope2StraightSupport': 213,
    'ArenaCheckpointUp': 214,
    'ArenaCheckpointUpX2': 215,
    'ArenaCheckpointDown': 216,
    'ArenaCheckpointDownX2': 217,
    'ArenaCheckpointLeft': 218,
    'ArenaCheckpointLeftX2': 219,
    'ArenaCheckpointRight': 220,
    'ArenaCheckpointRightX2': 221,
    'ArenaTurboUp': 222,
    'ArenaTurboDown': 223,
    'ArenaTurboLeft': 224,
    'ArenaTurboRight': 225,
    'ArenaTurbo2Up': 226,
    'ArenaTurbo2Down': 227,
    'ArenaTurbo2Left': 228,
    'ArenaTurbo2Right': 229,
    'ArenaFWUD': 230,
    'ArenaFWLR': 231,
    'ArenaLoopStart': 232,
    'ArenaLoopStraight2': 233,
    'ArenaLoopEnd': 234,
    'ArenaLoopStraight2Corner': 235,
    'ArenaLoopStraight2GTCurve2': 236,
    'ArenaBorderStraight': 237,
    'ArenaBorderStraightFlip': 238,
    'ArenaLoopCheckpointH': 239,
    'ArenaLoopCheckpointHX2': 240,
    'ArenaLoopCheckpointV': 241,
    'ArenaLoopCheckpointVX2': 242,
    'ArenaSupport': 243,
    'ArenaLoopSupport': 244,
    'ArenaLoopSupport2': 245,
    'ArenaLoopAdvert2x1': 246,
    'ArenaBorderStraight_Advert2x1': 247,
    'ArenaBorderStraightFlip_Advert2x1': 248,
    'RoadTunnelStraight': 249,
    'RoadTunnelCross': 250,
    'RoadTunnelCorner': 251,
    'RoadTunnelGTCurve2': 252,
    'RoadTunnelGTCurve3': 253,
    'RoadTunnelCheckpoint': 254,
    'RoadTunnelFinish': 255,
    'RoadTunnelTurbo': 256,
    'RoadTunnelTurbo2': 257,
    'RoadTunnelFW': 258,
    'RoadTunnelSlope2Base': 259,
    'RoadTunnelSlope2Base2x1': 260,
    'RoadTunnelSlope2Base3x1': 261,
    'RoadTunnelSlope2Start': 262,
    'RoadTunnelSlope2StartX2': 263,
    'RoadTunnelSlope2End': 264,
    'RoadTunnelSlope2EndX2': 265,
    'RoadTunnelSlope2UTop': 266,
    'RoadTunnelSlope2UBottom': 267,
    'RoadTunnelSlope2Base3x1_GT3Left': 268,
    'RoadTunnelSlope2Base3x1_GT3Right': 269,
    'RoadTunnelSlope2Straight': 270,
    'RoadTunnelCheckpointUp': 271,
    'RoadTunnelCheckpointDown': 272,
    'RoadTunnelTurboUp': 273,
    'RoadTunnelTurboDown': 274,
    'RoadTunnelTurbo2Up': 275,
    'RoadTunnelTurbo2Down': 276,
    'RoadTunnelFWSlope': 277,
    'HighWayTunnel Exit2x1Land': 278,
    'HighWayTunnelSlope2Exit2x1Land': 279,
    'HighWayTunnel EntryLandHill': 280,
    'HighWayTunnel EntryLandHill6': 281,
    'HighWayTunnelEntryLandHill6_2': 282,
    'HighWayTunnelEntryLandHill6_4': 283,
    'HighWayTunnel EntryLandCliff12': 284,
    'HighWayTunnelEntryLandCliff12_2': 285,
    'HighWayTunnelEntryLandCliff12_4': 286,
    'HighWayTunnelEntryLandCliff12_6': 287,
    'HighWayTunnelEntryLandCliff12_8': 288,
    'HighWayTunnelEntryLandCliff12_10': 289,
    'HighWayTunnelEntryLandSeaCliff13_1': 290,
    'HighWayTunnelEntryLandSeaCliff13_3': 291,
    'HighWayTunnelEntryLandSeaCliff13_5': 292,
    'HighWayTunnelEntryLandSeaCliff13_7': 293,
    'HighWayTunnelEntryLandSeaCliff13_9': 294,
    'HighWayTunnelEntryLandSeaCliff13_11': 295,
    'DecoTreeBeach': 296,
    'DecoBeachMangrove': 297,
    'DecoTreeBeachDeadend3x2': 298,
    'DecoTreeBeachDeadend3x2In': 299,
    'DecoTreeBeachGT2In': 300,
    'DecoTreeBeachGT2out': 301,
    'DecoTreeBeachGT3In': 302,
    'DecoBeachSandbar2x1Mirror': 303,
    'DecoBeachSandbar2x1': 304,
    'DecoBeachGTDiag3Mirror': 305,
    'DecoBeachGTDiag3': 306,
    'BeachGT2RacingOut': 307,
    'DecoBeachGTDiag3RacingMirror': 308,
    'DecoBeachGTDiag3Racing': 309,
    'DecoBeachBase5-2x1Mirror': 310,
    'DecoBeachBase5-2x1': 311,
    'DecoBeach-Ramp': 312,
    'DecoBeach-RampMirror': 313,
    'DecoBeach-Ramp2': 314,
    'DecoBeach-Ramp2Mirror': 315,
    'BeachCheckPointLeft': 316,
    'BeachCheckPointRight': 317,
    'FreeWayToBeachDiag3X2': 318,
    'FreeWayToBeachDiag3X2Mirror': 319,
    'FreeWayToBeachGTCurve3In': 320,
    'FreeWayToBeachGTCurve3InMirror': 321,
    'FreeWayToBeachGTCurve3Out': 322,
    'FreeWayToBeachGTCurve3OutMirror': 323,
    'DecoTreePalmBig': 324,
    'DecoTreeJungle': 325,
    'DecoTerraformingHD': 326,
    'FieldLandHill6GT3': 327,
    'DecoBirds': 328,
    'LandCheckpoint': 329,
    'LandHillCheckpointUp': 330,
    'LandHillCheckpointDown': 331,
    'LandHillCheckpointLeft': 332,
    'LandHillCheckpointRight': 333,
    'DecoHotelB_2x1': 334,
    'DecoHotelC_2x1': 335,
    'DecoHotelD_2x1': 336,
    'Resort2x2_InDecoHotelC': 337,
    'DecoResortBeach-Deadend': 338,
    'DecoResortBeach-DeadendPool': 339,
    'DecoCityPodium': 340,
    'PodiumResort': 341,
    'DecoCity_Support': 342,
    'DecoWheel': 343,
    'DecoWheelDiag': 344,
    'DecoCityFloorBase': 345,
    'LandToCityStraight': 346,
    'LandToCityCornerIn': 347,
    'LandToCityCornerOut': 348,
    'LandToCityBase5': 349,
    'DecoCityCheckpoint': 350,
    'DecoCityFloorTurbo': 351,
    'DecoCityFloorTurbo2': 352,
    'DecoCityFloorFW': 353,
    'RoadMainToCity': 354,
    'DecoCityFloorSlopeStraight': 355,
    'DecoCityFloorSlopeLiteStraight': 356,
    'DecoCityFloorSlopeCornerIn': 357,
    'DecoCityFloorSlopeLiteCornerIn': 358,
    'DecoCityFloorSlopeCornerOut': 359,
    'DecoCityFloorSlopeLiteCornerOut': 360,
    'DecoCityFloorRampCenter': 361,
    'DecoCityFloorRamp2Center': 362,
    'DecoCityGardenBase': 363,
    'DecoCityGardenCornerIn': 364,
    'DecoCityGardenCornerOut': 365,
    'DecoCityGardenBorderStraight': 366,
    'DecoCityGardenBorderCornerIn': 367,
    'DecoCityGardenBorderCornerOut': 368,
    'DecoCityGardenBorderRampStraight': 369,
    'DecoCityGardenRampCornerIn': 370,
    'DecoCityGardenSlopeStraight': 371,
    'DecoCityGardenSlopeLiteStraight': 372,
    'DecoCityGardenSlopeCornerIn': 373,
    'DecoCityGardenSlopeLiteCornerIn': 374,
    'DecoCityGardenSlopeCornerOut': 375,
    'DecoCityGardenSlopeLiteCornerOut': 376,
    'AdPanel': 377,
    'TechLight': 378,
    'CellTowerTall': 379,
    'CellTowerMedium': 380,
    'CellTowerSmall': 381,
    'DecoShipBack': 382,
    'DecoShipBackMirror': 383,
    'DecoShipFront': 384,
    'DecoShipFrontMirror': 385,
    'DecoShipMid': 386,
    'CE_ShipMidXBeachGT2': 387,
    'CE_ShipMidXBeachGT2Mirror': 388,
    'DecoShipMidContainerCenterMirror': 389,
    'DecoShipMidContainerCenter': 390,
    'DecoShipMidContainerDouble': 391,
    'DecoShipMidContainerDoubleMirror': 392,
    'CE_ShipMidContainerLeft': 393,
    'CE_ShipMidContainerLeftMirror': 394,
    'CE_ShipMidContainerRight': 395,
    'CE_ShipMidContainerRightMirror': 396,
}


VALLEY_BLOCKS = {
    'Road2Way': 1,
    'Road2WayStart': 2,
    'Road2WayCheckpoint': 3,
    'Road2WayFinish': 4,
    'Road2WayMultilap': 5,
    'Road2WayWithBirds': 6,
    'Road2WayWithoutPole': 7,
    'RailTrackXRoad2Way': 8,
    'Road2Way_Straight_InViaduc': 9,
    'Road2WayStraightSupport': 10,
    'Road2WayGTCurve2': 11,
    'Road2WayGTCurve3': 12,
    'Road2WayGTCurve4': 13,
    'Road2WayDiag2Mirror': 14,
    'Road2WayDiag2': 15,
    'Road2WayDiag3Mirror': 16,
    'Road2WayDiag3': 17,
    'Road2WayDiag4Mirror': 18,
    'Road2WayDiag4': 19,
    'Road2WayBiSlope': 20,
    'Road2WayBiSlopeStart': 21,
    'Road2WayCliff2Straight3': 22,
    'Road2WayCliff2Straight3_HillsLeft': 23,
    'Road2WayCliff2Straight3_HillsRight': 24,
    'Road2WaySlopeStraight': 25,
    'Road2WayBiSlopeEnd': 26,
    'Road2WaySlopeUTopAir': 27,
    'Road2WaySlopeUBottomGround': 28,
    'Road2WayCheckpointSlopeUp': 29,
    'Road2WayCheckpointSlopeDown': 30,
    'Road2WayCliff2Curve3_Left': 31,
    'Road2WayCliff2Curve3_Right': 32,
    'Road2WayCliff2Curve3In_HillsLeft': 33,
    'Road2WayCliff2Curve3In_HillsRight': 34,
    'Road2WayCliff2Curve3Out_HillsLeft': 35,
    'Road2WayCliff2Curve3Out_HillsRight': 36,
    'Road2WayGTCurve2Slope': 37,
    'Road2WayGTCurve2SlopeMirror': 38,
    'Road2WayGTCurve2SlopeStart': 39,
    'Road2WayGTCurve2SlopeStartMirror': 40,
    'Road2WayGTCurve2SlopeEnd': 41,
    'Road2WayGTCurve2SlopeEndMirror': 42,
    'Road2WayCliff4Curve3In_Left': 43,
    'Road2WayCliff4Curve3In_Right': 44,
    'Road2WayCliff4Curve3Out_Left': 45,
    'Road2WayCliff4Curve3Out_Right': 46,
    'Road2WayCliff4Side5_Left': 47,
    'Road2WayCliff4Side5_Right': 48,
    'Cliff8SlopeCorner_6x7x8': 49,
    'Cliff8SlopeCorner_6x7x8_Mirror': 50,
    'Cliff8SlopeDiag_8x3x8': 51,
    'Cliff8SlopeDiag_8x3x8_Mirror': 52,
    'Road2WayTetrisCliff3x2': 53,
    'Road2WayTetrisCliff3x2Mirror': 54,
    'Road2WaySlopeBase2x1OnHilll': 55,
    'Road2WaySlopeBaseGTCurve2OnHill': 56,
    'Road2WaySlopeBaseGTCurve2OnHillMirror': 57,
    'Road2WayTetrisHill3x2': 58,
    'Road2WayTetrisHill3x2Mirror': 59,
    'Road2WayBridgeGrassHills': 60,
    'Road2WayBridgeGrassCliff2': 61,
    'Road2WayBridgeGrassCliff4': 62,
    'Road2WayBridgeGrassCliff8': 63,
    'Road2WayTurbo': 64,
    'Road2WayTurboUp': 65,
    'Road2WayTurboDown': 66,
    'Road2WayFW': 67,
    'Road2WayFWSlope': 68,
    'Road4WayToRoad2WayStraight': 69,
    'Road4WayToRoad2WayTShaped': 70,
    'Road4WayCrossRoad2Way': 71,
    'Road4WayToRoad2WayCorner': 72,
    'Road4WayToRoad2WayCornerMirror': 73,
    'Road4WayToRoad2WayStraight2x1': 74,
    'Road4WayToRoad2WayStraight2x1Mirror': 75,
    'Road4WayToRoad2WayDiag': 76,
    'Road4WayToRoad2WayDiagMirror': 77,
    'Road4WayToRoad2WayGTCurve2': 78,
    'Road4WayToRoad2WayGTCurve2Mirror': 79,
    'Road2WayForestStraight': 80,
    'Road2WayForestCorner': 81,
    'Road2WayForestTShaped': 82,
    'Road2WayForestCross': 83,
    'Road2WayForestDeadend': 84,
    'Road2WayForestStraightNarrow': 85,
    'Road2WayForestStart': 86,
    'Road2WayForestCheckpoint': 87,
    'Road2WayForestFinish': 88,
    'Road2WayForestMultilap': 89,
    'Road2WayForestTurbo': 90,
    'Road2WayForestFW': 91,
    'Road2WayForestGTCurve2': 92,
    'Road2WayForestGTCurve3': 93,
    'Road2WayForestGTCurve3Field': 94,
    'Road2WayForestDiag2': 95,
    'Road2WayForestDiag2Mirror': 96,
    'Road2WayForestDiag3': 97,
    'Road2WayForestDiag3Mirror': 98,
    'Road2WayForestSlopeBaseOnHill': 99,
    'Road2WayForestSlopeBase2x1OnHill': 100,
    'Road2WayForestTetrisHill3x2': 101,
    'Road2WayForestTetrisHill3x2Mirror': 102,
    'Road2WayForestSlopeGTCurve2OnHillMirror': 103,
    'Road2WayForestSlopeGTCurve2OnHill': 104,
    'Road2WayForestStraightToRoadDirtLarge': 105,
    'Road2WayForestXRoadForestLarge': 106,
    'Road2WayForestStraightToRoad2Way': 107,
    'Road2WayForestSlopeStartOut': 108,
    'Road2WayForestSlopeBase2x1Out': 109,
    'Road4Way': 110,
    'Road4WayStart': 111,
    'Road4WayCheckpoint': 112,
    'Road4WayFinish': 113,
    'Road4WayMultiLap': 114,
    'Road4WayBumper': 115,
    'Road4WayStraightWithBirds': 116,
    'Road4WayStraightNoLights': 117,
    'HighwayStraightSupport': 118,
    'Road4WayGrassCliff2Straight3': 119,
    'Road4WayHillStraight2': 120,
    'Road4WaySlopeBase': 121,
    'Road4WaySlopeStraight': 122,
    'Road4wayBiSlopeStart': 123,
    'Road4WayBiSlopeEnd': 124,
    'Road4WayBiSlope': 125,
    'Road4WaySlopeUTop': 126,
    'Road4WaySlopeUBottom': 127,
    'Road4WayCheckpointUp': 128,
    'Road4WayCheckpointDown': 129,
    'Road4WayGTCurve2': 130,
    'Road4WayGTCurve3': 131,
    'Road4wayGTCurve4': 132,
    'Road4wayDiag3': 133,
    'Road4wayDiag3Mirror': 134,
    'Road4wayDiag4MirrorRoad4wayDiag4': 135,
    'Road4wayDiag4': 136,
    'Road4WayHillCurve2': 137,
    'Road4WayHillCurve2Mirror': 138,
    'Road4WayGTCurve2SlopeMirror': 139,
    'Road4WayGTCurve2Slope': 140,
    'Road4WayGTCurve3Slope': 141,
    'Road4WayGTCurve3SlopeMirror': 142,
    'Road4WayLooping': 143,
    'Road4WayLoopingMirror': 144,
    'Road4WayFreeLoopStart': 145,
    'Road4WayFreeLoopStraight': 146,
    'Road4WayFreeLoopEnd': 147,
    'Road4WayBridgeGrassHills': 148,
    'Road4WayBridgeGrassCliff2': 149,
    'Road4WayBridgeGrassCliff4': 150,
    'Road4WayBridgeGrassCliff8': 151,
    'Road4WayTurbo': 152,
    'Road4WayTurboUp': 153,
    'Road4WayTurboDown': 154,
    'Road4WayFW': 155,
    'Road4WaySlopeFW': 156,
    'Road4WayToHighway': 157,
    'Road4WayToHighwayNoSign': 158,
    'Road4WayToHighwayLeft': 159,
    'Road4WayToHighwayRight': 160,
    'HighwayStraight': 161,
    'HighwayStraightNoLights': 162,
    'HighwaySwitch': 163,
    'HighwaySwitchLeft': 164,
    'HighwaySwitchRight': 165,
    'HighwayGTCurve4': 166,
    'HighwayBiSlopeStart': 167,
    'HighwaySlopeStraight': 168,
    'HighwayBiSlopeEnd': 169,
    'HighwayBridgeGrassHills': 170,
    'HighwayBridgeGrassCliff2': 171,
    'HighwayBridgeGrassCliff4': 172,
    'HighwayBridgeGrassCliff8': 173,
    'TunnelRoad4wayStraight': 174,
    'TunnelRoad4wayCross': 175,
    'TunnelRoad4wayCorner': 176,
    'TunnelRoad4wayGTCurve2': 177,
    'TunnelRoad4wayGTCurve3': 178,
    'TunnelRoad4wayCheckpoint': 179,
    'TunnelRoad4wayFinish': 180,
    'TunnelRoad4WayTurbo': 181,
    'TunnelRoad4WayFW': 182,
    'TunnelRoad4waySlopeBase': 183,
    'TunnelRoad4waySlopeBaseX2': 184,
    'TunnelRoad4wayGTCurve3SlopeMirror': 185,
    'TunnelRoad4wayGTCurve3Slope': 186,
    'TunnelRoad4wayBiSlopeStart': 187,
    'TunnelRoad4waySlopeStraight': 188,
    'TunnelRoad4wayBiSlopeEnd': 189,
    'Road4wayUTop': 190,
    'Road4wayUbottom': 191,
    'TunnelRoad4WayCheckpointUp': 192,
    'TunnelRoad4WayCheckpointDown': 193,
    'TunnelRoad4WayTurboUp': 194,
    'TunnelRoad4WayTurboDown': 195,
    'TunnelRoad4WayFWUD': 196,
    'Road4WayEntryGrassSlopeBase2': 197,
    'TunnelEntryRoad4WayGrassCliff2': 198,
    'TunnelEntryRoad4WayGrassCliff4': 199,
    'TunnelEntryRoad4WayGrassCliff4_1': 200,
    'TunnelEntryRoad4WayGrassCliff8': 201,
    'TunnelEntryRoad4WayGrassCliff8_1': 202,
    'TunnelEntryRoad4WayGrassCliff8_2': 203,
    'TunnelEntryRoad4WayGrassCliff8_3': 204,
    'TunnelStraight': 205,
    'TunnelHighwaySwitch': 206,
    'TunnelHighwaySwitchLeft': 207,
    'TunnelHighwaySwitchRight': 208,
    'TunnelHighwayGTCurve2': 209,
    'TunnelHighwayGTCurve3': 210,
    'TunnelGTCurve4': 211,
    'TunnelRoad4wayToHighway': 212,
    'TunnelRoad4WayToHighWayLeft': 213,
    'TunnelRoad4WayToHighWayRight': 214,
    'TunnelHighwayTurbo': 215,
    'TunnelHighwayFW': 216,
    'HighwayEntryGrassSlopeBase2': 217,
    'TunnelEntryGrassCliff2': 218,
    'TunnelEntryGrassCliff4': 219,
    'TunnelEntryGrassCliff4_1': 220,
    'TunnelEntryGrassCliff8': 221,
    'TunnelEntryGrassCliff8_1': 222,
    'TunnelEntryGrassCliff8_2': 223,
    'TunnelEntryGrassCliff8_3': 224,
    'ArenaParkRoofBase': 225,
    'ArenaParkRoofBaseSupport': 226,
    'ArenaParkRoofBaseNoLights': 227,
    'ArenaCheckpointParkRoofBaseAir': 228,
    'ArenaMultilapParkRoofBase': 229,
    'ArenaParkRoofTurbo': 230,
    'ArenaParkRoofFW': 231,
    'ArenaBaseSupport': 232,
    'ArenaBase': 233,
    'ArenaBaseCornerIn': 234,
    'ArenaBaseSlopeBase': 235,
    'ArenaBaseSlopeStart': 236,
    'ArenaBaseSlopeStraight': 237,
    'ArenaBaseSlopeEnd': 238,
    'ArenaParkRoofSlopeBase': 239,
    'ArenaParkRoofSlopeStart': 240,
    'ArenaParkRoofSlopeStraight': 241,
    'ArenaParkRoofSlopeEnd': 242,
    'ArenaCheckpointParkRoofSlopeStraightUpAir': 243,
    'ArenaCheckpointParkRoofSlopeStraightDownAir': 244,
    'ArenaCheckpointParkRoofSlopeLeft': 245,
    'ArenaCheckpointParkRoofSlopeRight': 246,
    'ArenaParkRoofTurboSlopeUp': 247,
    'ArenaParkRoofTurboSlopeDown': 248,
    'ArenaParkRoofTurboSlopeLeft': 249,
    'ArenaParkRoofTurboSlopeRight': 250,
    'ArenaParkRoofFWSlopeUD': 251,
    'ArenaParkRoofFWSlopeLR': 252,
    'ArenaBorderStraight': 253,
    'ArenaBorderStraightArch': 254,
    'ArenaBorderSmallLoopStraightTop': 255,
    'ArenaBorderCornerIn': 256,
    'ArenaBorderGTCurve2In': 257,
    'ArenaBorderGTCurve3In': 258,
    'ArenaBorderCornerOut': 259,
    'ArenaBorderGTCurve2Out': 260,
    'ArenaBorderGTCurve3Out': 261,
    'ArenaSupport': 262,
    'ArenaBorderSmallStraight': 263,
    'ArenaBorderSmallStraightOpen': 264,
    'ArenaBorderSmallStraightOpenFence': 265,
    'ArenaBorderSmallCornerIn': 266,
    'ArenaBorderSmallGTCurve2In': 267,
    'ArenaBorderSmallGTCurve3In': 268,
    'ArenaBorderSmallCornerOut': 269,
    'ArenaBorderSmallGTCurve2Out': 270,
    'ArenaBorderSmallGTCurve3Out': 271,
    'ArenaBorderSmallSlopeBaseLeft': 272,
    'ArenaBorderSmallSlopeBaseRight': 273,
    'ArenaBorderSmallSlopeStartLeft': 274,
    'ArenaBorderSmallSlopeStartRight': 275,
    'ArenaBorderSmallSlopeStraightLeft': 276,
    'ArenaBorderSmallSlopeStraightRight': 277,
    'ArenaBorderSmallSlopeEndLeft': 278,
    'ArenaBorderSmallSlopeEndRight': 279,
    'ArenaParkBase': 280,
    'ArenaParkBaseGrassCenter': 281,
    'ArenaParkSlopeBase': 282,
    'ArenaBaseToParkUp': 283,
    'ArenaBaseToParkDown': 284,
    'ArenaParkBorderStraight': 285,
    'ArenaParkBorderStraightOpen': 286,
    'ArenaParkBorderGTCurve2In': 287,
    'ArenaParkBorderGTCurve3In': 288,
    'ArenaParkBorderCornerOut': 289,
    'ArenaParkBorderGTCurve2Out': 290,
    'ArenaParkBorderGTCurve3Out': 291,
    'ArenaParkRoofLoopStartSmall': 292,
    'ArenaBorderSmallLoopSmallStartRight': 293,
    'ArenaBorderSmallLoopSmallStartLeft': 294,
    'ArenaParkRoofLoopStart': 295,
    'ArenaBorderSmallLoopStartLeft': 296,
    'ArenaBorderSmallLoopStartRight': 297,
    'ArenaParkRoofLoopEndSmall': 298,
    'ArenaBorderSmallLoopSmallEndRight': 299,
    'ArenaBorderSmallLoopSmallEndLeft': 300,
    'ArenaParkRoofLoopEnd': 301,
    'ArenaBorderSmallLoopEndRight': 302,
    'ArenaBorderSmallLoopEndLeft': 303,
    'ArenaParkRoofLoopStraight1': 304,
    'ArenaParkRoofLoopStraight2': 305,
    'ArenaParkRoofLoopStraight4': 306,
    'ArenaCheckpointParkRoofLoopStraight2VAir': 307,
    'ArenaCheckpointParkRoofLoopStraight4HAir': 308,
    'ArenaBorderSmallLoopStraightLeft': 309,
    'ArenaBorderSmallLoopStraightRight': 310,
    'ArenaParkRoofLoop2GtCurve2': 311,
    'ArenaParkRoofLoop2GtCurve2ToPark': 312,
    'ArenaParkRoofLoop2GtCurve2ToRubber': 313,
    'ArenaScreenParkRoofLoopStraight1': 314,
    'ArenaScreenBaseX1': 315,
    'ArenaScreenBaseX4': 316,
    'ArenaScreenBorderSmallStraightOpen': 317,
    'ArenaBorderSmallStraight': 318,
    'ArenaBaseToRoad4Way': 319,
    'ArenaParkToRoad4Way': 320,
    'ArenaParkRoofToRoad4Way': 321,
    'ArenaBorderToRoad4Way': 322,
    'RoadForestLargeStraight': 323,
    'RoadForestLargeDeadend': 324,
    'RoadForestLargeStart': 325,
    'RoadForestLargeCheckpoint': 326,
    'RoadForestLargeFinish': 327,
    'RoadForestLargeMultilap': 328,
    'Road2WayXRoadForestLarge': 329,
    'RailTrackXRoadDirtLarge': 330,
    'RoadForestLargeToRoadDirt': 331,
    'RoadForestLargeCorner': 332,
    'RoadForestLargeCornerSignBarrier': 333,
    'RoadForestLargeGTCurve2': 334,
    'RoadForestLargeCross': 335,
    'RoadForestLargeDiag2': 336,
    'RoadForestLargeDiag2Mirror': 337,
    'RoadForestLargeSlope': 338,
    'RoadForestLargeSlopeX2': 339,
    'RoadForestLargeGTCurve2SlopeMirror': 340,
    'RoadForestLargeGTCurve2Slope': 341,
    'RoadForestLargeSlopeCornerUturn': 342,
    'RoadForestLargeStartLoop': 343,
    'RoadForestLargeTurbo': 344,
    'RoadForestLargeFW': 345,
    'RoadForestLargeBump': 346,
    'RoadForestLargeStraightBarrier': 347,
    'RoadForestLargeStraightObstacle': 348,
    'RoadForestLargeStraightNarrow': 349,
    'RoadForestLargeStraightWithBirds': 350,
    'RoadDirtCornerSign': 351,
    'RoadDirtGTCurve2': 352,
    'RoadDirtDiag2x2Mirror': 353,
    'RoadDirtDiag2x2': 354,
    'RoadDirtCheckpoint': 355,
    'Road2WayToRoadDirt': 356,
    'RoadDirtToFieldEntry': 357,
    'RoadDirt': 358,
    'RoadDirtNoTree': 359,
    'RoadDirt_Straight_InViaducPillar': 360,
    'RoadDirtRiverCrossStraight': 361,
    'RoadDirtRiverCrossDeepStraightx2Mirror': 362,
    'RoadDirtRiverCrossDeepStraightx2': 363,
    'RoadDirtGTCurve2Pond': 364,
    'RoadDirtGTCurve2Lake': 365,
    'RoadDirtDiag2x2PondMirror': 366,
    'RoadDirtDiag2x2Pond': 367,
    'RoadDirtDiag2x2LakeMirror': 368,
    'RoadDirtDiag2x2Lake': 369,
    'RoadDirtRiverCrossDiag3x2': 370,
    'RoadDirtRiverCrossDiag3x2Mirror': 371,
    'RoadDirtDeadendLakeJump2': 372,
    'RoadDirtDeadendLakeJump': 373,
    'RoadCliffSlopeStart': 374,
    'RoadCliffSlopeStartMirror': 375,
    'RoadCliffSlopeStartGTCurveOut': 376,
    'RoadCliffSlopeStartGTCurveOutMirror': 377,
    'RoadCliffSlopeStartGTCurveIn': 378,
    'RoadCliffSlopeStartGTCurveInMirror': 379,
    'RoadCliffSlopeEnd': 380,
    'RoadCliffSlopeEndMirror': 381,
    'RoadCliffSlopeEndGTCurveOut': 382,
    'RoadCliffSlopeEndGTCurveOutMirror': 383,
    'RoadCliffSlopeEndGTCurveIn': 384,
    'RoadCliffSlopeEndGTCurveInMirror': 385,
    'RoadCliffSlopeUTop': 386,
    'RoadCliffSlopeUTopx2': 387,
    'RoadCliffSlopeUBottom': 388,
    'RoadCliffSlopeUBottomx2': 389,
    'RoadCliffStraightBumpRight': 390,
    'RoadCliffStraightBumpLeft': 391,
    'RoadCliffUBottomGT2In': 392,
    'RoadCliffUBottomGT2Out': 393,
    'RoadCliffRoadDirtLargeArch2x1': 394,
    'RoadCliffSlopeUTopCheckPointLeft': 395,
    'RoadCliffSlopeUTopCheckPointRight': 396,
    'RoadCliffSlopeUBottomCheckPointLeft': 397,
    'RoadCliffSlopeUBottomCheckPointRight': 398,
    'RoadCliffTetrisSlopeStraight3x3': 399,
    'RoadCliffTetrisCliff3x2': 400,
    'RoadCliffTetrisCliff3x2Mirror': 401,
    'RoadCliff2TetrisGTCurve2': 402,
    'RoadCliff2TetrisGTCurve2Mirror': 403,
    'DecoCastleDeadendEntryLake': 404,
    'DecoCastleDeadendEntryRacingLake': 405,
    'DecoCastleEntryCheckPointOut': 406,
    'DecoCastleEntryCheckPointIn': 407,
    'DecoCastleBridgeLake': 408,
    'DecoCastleBridgeToRoad2WayLake': 409,
    'DecoCastleDeadendWallLake': 410,
    'DecoCastleDeadendWallRacingLake': 411,
    'DecoCastleDeadendTowerHouseLake': 412,
    'DecoCastleDeadendTowerHouseLakeMirror': 413,
    'DecoCastleDeadendWallCheckPointLeft': 414,
    'DecoCastleDeadendWallCheckPointRight': 415,
    'DecoCastleDeadendWallRacingCheckPointLeft': 416,
    'DecoCastleDeadendWallRacingCheckPointRight': 417,
    'DecoCastleCornerWallLake': 418,
    'DecoCastleCornerWallRacingLake': 419,
    'DecoCastleCornerTowerHouseRacingLake': 420,
    'DecoCastleCornerTowerHouseLake': 421,
    'DecoCastleGTCurve2WallRacingLake': 422,
    'DecoCastleCornerInWallLake': 423,
    'DecoCastleCornerInWallRacingLake': 424,
    'DecoCastleGTCurve2InWallRacingLake': 425,
    'DecoCastleDeadendWallSlopeLakeRight': 426,
    'DecoCastleDeadendWallSlopeLakeLeft': 427,
    'DecoCastleCenterLake': 428,
    'DecoCastleCenterCornerSignLake': 429,
    'DecoCastleCenterCheckpointLake': 430,
    'DecoCastleCenterMultilapLake': 431,
    'DecoCastleGardenFlowerStraightHalfLake': 432,
    'DecoCastleGardenFlowerCornerHalfOutLake': 433,
    'DecoCastleGardenFlowerCornerHalfInLake': 434,
    'DecoCastleGardenFlowerStraightHalfCheckpointLakeLeft': 435,
    'DecoCastleGardenFlowerStraightHalfCheckpointLakeRight': 436,
    'DecoCastleGardenFlowerStraightx2CheckpointLake': 437,
    'DecoCastleGardenStraightLake': 438,
    'DecoCastleGardenFlowerStraightLake': 439,
    'DecoCastleGardenCornerLake': 440,
    'DecoCastleGardenFlowerCornerLake': 441,
    'DecoCastleGardenCornerSignLake': 442,
    'DecoCastleGardenFlowerCornerSignLake': 443,
    'DecoCastleGardenDeadendLake': 444,
    'DecoCastleGardenFlowerDeadendLake': 445,
    'DecoCastleGardenCheckpointLake': 446,
    'DecoCastleGardenMultilapLake': 447,
    'ViaducStraight': 448,
    'ViaducArchStraightX3': 449,
    'ViaducArchStraightX5': 450,
    'ViaducCurve2': 451,
    'ViaducCurve3': 452,
    'ViaducCurve4': 453,
    'Viaduc_Straight_InViaduc': 454,
    'ViaducCheckpoint': 455,
    'Viaduc_SlopeBase_Air': 456,
    'Viaduc_Slope2Start_Air': 457,
    'Viaduc_Slope2Straight_Air': 458,
    'Viaduc_Slope2End_Air': 459,
    'Viaduc_SlopeBase2x1_Air': 460,
    'Viaduc_Slope2Start2x2x1_Air': 461,
    'ViaducToRoad2Way_Slope2Start2x2x1_Ground': 462,
    'ViaducToRoadForestLarge_Slope2Start2x2x1_Ground': 463,
    'Viaduc_Slope2End2x2x1_Air': 464,
    'Viaduc_Slope2UTop_Air': 465,
    'Viaduc_Slope2UBottom_Air': 466,
    'ViaducCheckpointSlopeUp': 467,
    'ViaducCheckpointSlopeDown': 468,
    'Viaduc_ToRoad2Way_Air': 469,
    'ViaducToRoad2WayGrassCliff2': 470,
    'ViaducToRoad2WayGrassCliff4': 471,
    'ViaducToRoad2WayGrassCliff8': 472,
    'ViaducToRoadDirtGrassCliff2': 473,
    'ViaducToRoadDirtGrassCliff4': 474,
    'ViaducToRoadDirtGrassCliff8': 475,
    'ViaducToRoadForestLargeGrassCliff2': 476,
    'ViaducToRoadForestLargeGrassCliff4': 477,
    'ViaducToRoadForestLargeGrassCliff8': 478,
    'RailTrackStraight': 479,
    'RailTrackSlopeBase6x1': 480,
    'RailTrackSlopeEnd6x1': 481,
    'RailTrackSlopeStart6x1': 482,
    'RailTrackSlopeStraight2x1': 483,
    'RailTrackSlopeBase3X1_GrassHill': 484,
    'RailTrackSlope2Base3X1_GrassCliff2': 485,
    'RailTrackBridgeGrassCliff2': 486,
    'RailTrackBridgeGrassCliff4': 487,
    'RailTrackBridgeGrassCliff8': 488,
    'RailTrackGTCurve2': 489,
    'RailTrackGTCurve3': 490,
    'RailTrackGTCurve4': 491,
    'RailTrackGTDiag2': 492,
    'RailTrackGTDiag2_Mirror': 493,
    'RailTrackGTDiag3': 494,
    'RailTrackGTDiag3_Mirror': 495,
    'RailTrackGTDiag4': 496,
    'RailTrackGTDiag4_Mirror': 497,
    'RailTunnel_Straight': 498,
    'RailTunnel_CheckPoint': 499,
    'RailTunnel_SlopeBase3x1': 500,
    'RailTunnel_SlopeEnd3x1': 501,
    'RailTunnel_SlopeStart3x1': 502,
    'RailTunnel_SlopeStraight2x1': 503,
    'RailTrackTunnelEntryGrassCliff2': 504,
    'RailTrackTunnelEntryGrassCliff4': 505,
    'RailTrackTunnelEntryGrassCliff8': 506,
    'RailTunnel_GTCurve2': 507,
    'RailTunnel_GTCurve3': 508,
    'RailTunnel_GTCurve4': 509,
    'RailTunnel_GTDiag2': 510,
    'RailTunnel_GTDiag2_MirrorRailTunnel_GTDiag2': 511,
    'RailTunnel_GTDiag3': 512,
    'RailTunnel_GTDiag3_Mirror': 513,
    'RailTunnel_GTDiag4': 514,
    'RailTunnel_GTDiag4_Mirror': 515,
    'DecoTreesSmall': 516,
    'DecoTreesMed': 517,
    'DecoTrees': 518,
    'DecoTerrainHD': 519,
    'Cliff8HillCorner_6x7x8': 520,
    'DecoCliff8_Arch_2x1': 521,
    'DecoBirds': 522,
    'OffRoadCheckpoint': 523,
    'OffRoadSignLeft': 524,
    'OffRoadSignRight': 525,
    'DecoFieldDeadendEntry1x1': 526,
    'DecoFieldDeadendEntry1x1B': 527,
    'DecoFieldDeadendEntry2x1': 528,
    'DecoFieldDeadendEntry2x1B': 529,
    'DecoFieldStraw': 530,
    'DecoFieldCornerEntry1x1': 531,
    'TechnicsLightsFront': 532,
    'PowerLinesStraight1x4': 533,
    'PowerLinesStraight1x4x2': 534,
    'PowerLinesStraight4x4': 535,
    'DecoAeolian': 536,
    'TechnicsScreenBig1x2Front': 537,
    'TechnicsScreen1x2': 538,
    'TechnicsScreen1x2Mirror': 539,
    'TechnicsScreenLong1x2': 540,
    'PodiumDirt': 541,
    'PodiumParking': 542,
    'DecoFarmA': 543,
    'DecoFarmB': 544,
    'DecoFarmC': 545,
    'DecoFarmD': 546,
    'DecoFarmE': 547,
    'DecoFarmF': 548,
    'DecoFarmG': 549,
    'DecoChurch': 550,
    'DecoMayorHouse': 551,
    'DecoHomeBig': 552,
    'DecoHomeBigB': 553,
    'DecoHomeSmall': 554,
    'DecoOldBusStop': 555,
    'DecoWaterBridge': 556,
    'DecoWaterMill': 557,
    'DecoHomeSmallSlopeA': 558,
}


from distutils.core import setup
setup(
  name = 'pygbx',
  packages = ['pygbx'],
  version = '0.3',
  license='GPL3',
  description = 'A Python library to parse GBX files',
  author = 'Adam Bieńkowski',              
  author_email = 'donadigos159@gmail.com',
  url = 'https://github.com/donadigo/pygbx',
  download_url = 'https://github.com/donadigo/pygbx/archive/0.3.zip',
  keywords = ['GBX', 'parser', 'TrackMania'],
  install_requires=[
          'python-lzo',
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)